/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.util.*;
import java.io.*;
import java.net.URI;
import java.util.regex.*;
import java.nio.file.*;
import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;
import java.util.zip.*;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.*;
import javax.swing.tree.*;

import com.dickimawbooks.texparserlib.*;
import com.dickimawbooks.texparserlib.aux.*;
import com.dickimawbooks.texparserlib.html.*;
import com.dickimawbooks.texparserlib.latex.*;

public class Book extends BookDocument implements ActionListener
{
   public Book(MakeJmlrBookGUI application, File mainFile)
     throws IOException
   {
      super(application, null, mainFile.getName());

      BookRoot root = new BookRoot(application, mainFile.getParentFile());
      root.book = this;

      parent = root;
      root.addChild(this);

      importFiles = new Vector<BookDocument>();

      initFaqUri();

      parent.populate(app.getBookContentsFilter());

      bookNode = new BookNode(this);
   }

   public Book(MakeJmlrBookGUI application, BookRoot root, File mainFile)
    throws IOException
   {
      super(application, root, mainFile.getName());

      root.addChild(this);
      root.book = this;

      importFiles = new Vector<BookDocument>();

      initFaqUri();

      bookNode = new BookNode(this);
   }

   public void populate()
   {
      getRoot().populate(app.getBookContentsFilter());
   }

   public void refreshBook()
      throws IOException
   {
      // Do all the import files still exist?

      for (int i = importFiles.size()-1; i >= 0; i--)
      {
         BookFile importFile = importFiles.get(i);

         File file = importFile.getFile();

         if (!file.exists())
         {
            importFile.parent.removeChild(importFile);
            importFiles.remove(i);
         }
      }

      // Remove non-existent files from parent

      parent.prune();

      // Repopulate

      parent.populate(app.getBookContentsFilter());

      app.refreshFileView();
      app.setInfo(app.getLabel("message.done"));
   }

   private static void initFaqUri()
   {
      if (proceedingsFaq == null)
      {
         proceedingsFaq = URI.create(MakeJmlrBookGUI.getLabel("uri.jmlrfaq"));
      }
   }

   public void buildPapersMenu(JMenu menu)
   {
      if (menu.getMenuComponentCount() > 0)
      {
         menu.removeAll();
      }

      for (int i = 0, n=importFiles.size(); i < n; i++)
      {
         BookDocument doc = importFiles.get(i);

         JMenuItem item = new JMenuItem(doc.getBaseName());
         item.setToolTipText(doc.getFile().getAbsolutePath());
         item.setActionCommand(""+i);
         item.addActionListener(this);

         menu.add(item);
      }
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      try
      {
         int i = Integer.parseInt(action);

         Thread thread = new ProcessPaperThread(app, importFiles.get(i));
         app.startWorkerThread(thread);
      }
      catch (NumberFormatException e)
      {
      }
   }

   public void clearAllErrorMessages()
   {
      clearErrorMessages();

      if (importFiles != null)
      {
         for (int i = 0, n = importFiles.size(); i < n; i++)
         {
            importFiles.get(i).clearErrorMessages();
         }
      }

      if (preface != null)
      {
         preface.clearErrorMessages();
      }
   }

   public boolean isRunRequired(boolean printVersion, long pdfMod)
   {
      if (super.isRunRequired(printVersion, pdfMod))
      {
         return true;
      }

      String prefSource = getPrefaceSource();

      long prefMod = 0L;

      if (prefSource != null)
      {
         prefMod = (new File(getLocation(), prefSource)).lastModified();
      }

      if (prefMod > pdfMod)
      {
         if (app.inDebugMode())
         {
            app.printAndLogMessage("Run required: preface file modified since pdf modified.\n"
             +"\n  pref mod: "+(new Date(prefMod))
             +"\n  pdf mod: "+(new Date(pdfMod)));
         }

         return true;
      }

      for (int i = 0, n = importFiles.size(); i < n; i++)
      {
         if (importFiles.get(i).isRunRequired(false, pdfMod))
         {
            return true;
         }
      }

      // walk through directory tree to see if there are any .tex
      // files that have been modified since pdfMod

      TeXModWalkListener listener = new TeXModWalkListener(pdfMod);

      getRoot().walk(listener);

      return listener.isRunRequired();
   }

   public void processBook()
      throws IOException,InterruptedException,InvalidBookException
   {
      processBook(false, false);
   }

   public void clearAux()
   {
      super.clearAux();

      for (int i = 0, n = importFiles.size(); i < n; i++)
      {
         importFiles.get(i).clearAux();
      }
   }

   public void addAuxFiles(Vector<File> auxList)
   {
      super.addAuxFiles(auxList);

      for (int i = 0, n = importFiles.size(); i < n; i++)
      {
         importFiles.get(i).addAuxFiles(auxList);
      }
   }

   public String getHtmlShortTitle()
   {
      return shortTitle == null ?  getHtmlTitle() : shortTitle.getHtml();
   }

   public String getHtmlAbbrvTitle()
   {
      String title = (shortTitle == null ?
             getHtmlTitle() : shortTitle.getHtml());

      if (jmlrVolume > 0)
      {
         title = String.format("%s Volume %d: %s",
          isWCP() ? "JMLR W&amp;CP" : "PMLR", jmlrVolume, title);
      }

      return title;
   }

   public void processBook(
      boolean clearImportAux, boolean printVersion)
      throws IOException,InterruptedException,InvalidBookException
   {
      processBook(clearImportAux, printVersion, isRunRequired(printVersion));
   }

   public void processBook(
      boolean clearImportAux, boolean printVersion, boolean doRun)
      throws IOException,InterruptedException,InvalidBookException
   {
      processBook(clearImportAux, printVersion, doRun, true);
   }

   private void processBook(
      boolean clearImportAux, boolean printVersion, boolean doRun, boolean checkBMKentry)
      throws IOException,InterruptedException,InvalidBookException
   {
      app.setInfo(app.getLabelWithValue("message.processing", getBaseName()));
      clearAllErrorMessages();

      PreambleParser preambleParser = new PreambleParser(new BookTeXApp(app));
      (new TeXParser(preambleParser)).parse(getFile());

      clsFile = preambleParser.getDocumentClass();

      ControlSequence logoCs = preambleParser.getControlSequence("@logo");
      ControlSequence logoUrlCs = preambleParser.getControlSequence("@logourl");

      if (logoCs != null && logoCs instanceof GenericCommand)
      {
         setLogo(((GenericCommand)logoCs).getDefinition());
      }

      if (logoUrlCs != null && logoUrlCs instanceof GenericCommand)
      {
         setLogoUrl(((GenericCommand)logoUrlCs).getDefinition());
      }

      if (clsFile == null)
      {
         throw new InvalidBookException(
          app.getLabelWithValue("error.missing_cls", "jmlrbook"));
      }

      if (!clsFile.getName().equals("jmlrbook"))
      {
         if (clsFile.getName().equals("jmlr"))
         {
            throw new InvalidBookException(
             app.getLabelWithValue("error.jmlr_not_jmlrbook", getFile().getName()));
         }
         else
         {
            throw new InvalidBookException(
               app.getLabelWithValues("error.wrong_cls", 
               clsFile.getName(), "jmlrbook"));
         }
      }

      File iccFile = null;

      if (printVersion)
      {
         iccFile = app.getICCFile();
         app.debug("ICC file: "+(iccFile==null?"not set":iccFile.getAbsolutePath()));
      }

      app.debug("updating main info");
      updateMainInfo();

      app.debug("checking for interrupt");
      app.checkForInterrupt();

      if (clearImportAux)
      {
         app.debug("clearing aux");
         for (int i = 0, n = importFiles.size(); i < n; i++)
         {
            app.checkForInterrupt();

            importFiles.get(i).clearAux();
         }

         doRun = true;
      }

      String jobname = null;

      if (printVersion)
      {
         jobname = getBaseName()+"-print";
      }

      app.debug("checking for interrupt");
      app.checkForInterrupt();

      if (doRun)
      {
         clearPackageList();

         latex(jobname, printVersion ? "\\def\\jmlrgrayscale{1}" : "");

         app.checkForInterrupt();
      }

      File prefFile = (preface == null ? null : preface.getDocFile());

      if (!isCiML() && prefFile != null && prefFile.exists()
        && !preface.getSourceFile().equals(getFile()))
      {
         preface.latex();
      }

      app.debug("Process import files (doRun:"+doRun+")");
      processImportFiles(doRun);

      // Make sure articles all have the correct page numbers

      if (doRun)
      {
         for (int i = 0, n = importFiles.size(); i < n; i++)
         {
            app.checkForInterrupt();
            BookDocument importFile = importFiles.get(i);
         }
      }

      if (doRun)
      {
         doRun = isRunRequired(printVersion);
      }

      app.debug("Process print:"+printVersion+" doRun:"+doRun);

      process(printVersion, doRun);

      if (clearAuxAndRerun && checkBMKentry)
      {
         app.debug("Possible BKM@entry confusion detected. Clearing aux and reprocessing");

         clearAux();

         processBook(clearImportAux, printVersion, doRun, false);
         return;
      }

      int rerunCount = 0;

      while (isRunRequired(printVersion) && rerunCount <= MakeJmlrBookGUI.MAX_RERUN)
      {
         app.checkForInterrupt();
         rerunCount++;
         app.debug("Run required");
         latex(jobname, printVersion ? "\\def\\jmlrgrayscale{1}" : "");
      }

      if (app.inDebugMode())
      {
         if (isRunRequired(printVersion))
         {
            app.printAndLogMessage("Run still required: rerunCount="+rerunCount);
         }
      }

      app.beepIfEnabled();

      String mess = getAllErrorMessages();

      if (mess != null)
      {
         app.error(app.getLabel("error.detected"));
      }

      if (iccFile != null)
      {
         app.checkForInterrupt();
         app.debug("Calling toPdfX");
         app.toPdfX(getTeXShortAuthor(), 
            getTeXShortTitle(), "none", iccFile,
            getPdfFile(printVersion));
      }

      app.checkForInterrupt();
      checkLoadedPackages();

      app.checkForInterrupt();
      app.updateMainPanel();

      app.checkForInterrupt();
      checkConforms(printVersion);
      app.checkForInterrupt();

   }

   public int latex(String jobname, String predef)
      throws IOException,InterruptedException
   {
      boolean orgUseFileLine = getApplication().getProperties().useFileLine();

      try
      {
         getApplication().getProperties().setUseFileLine(true);

         return super.latex(jobname, predef);
      }
      finally
      {
         getApplication().getProperties().setUseFileLine(orgUseFileLine);
      }
   }

   public void checkLoadedPackages()
   {
      for (BookDocument doc : importFiles)
      {
         checkLoadedPackages(doc);
      }
   }

   public void checkLoadedPackages(BookDocument doc)
   {
      for (LaTeXFile userSty : doc.getUserPackages())
      {
         String name = userSty.getName();
         LaTeXFile mainSty = getUserPackage(name);

         Calendar cal = getPackageDate(name);

         // If cal != null && mainSty == null,
         // then package has been loaded implicitly

         if (mainSty == null && cal == null)
         {
            if (isProblemPackage(name))
            {
               addErrorMessage(app.getLabelWithValues("error.book.problem_sty",
                 new String[]
                 {
                    doc.getBaseName(),
                    name
                 })
               );
            }
            else
            {
               addErrorMessage(app.getLabelWithValues("error.book.missing_sty",
               new String[]
               {
                  doc.getBaseName(),
                  name
               })
               +getInformationLink("missing_sty:"+name+":"
                  +doc.getFile()));
            }

            continue;
         }

         Calendar docCal = doc.getPackageDate(name);

         if (cal == null || docCal == null)
         {
            continue;
         }
         else
         {

            int year = cal.get(Calendar.YEAR);
            int month = cal.get(Calendar.MONTH);
            int day = cal.get(Calendar.DAY_OF_MONTH);

            int docyear = docCal.get(Calendar.YEAR);
            int docmonth = docCal.get(Calendar.MONTH);
            int docday = docCal.get(Calendar.DAY_OF_MONTH);

            if (!(year == docyear && month == docmonth && day == docday))
            {
               addErrorMessage(app.getLabelWithValues("error.book.mismatched_sty",
               new String[]
               {
                  doc.getBaseName(),
                  name,
                  ""+docyear+"/"+docmonth+"/"+docday,
                  ""+year+"/"+month+"/"+day
               })
               +getInformationLink("mismatched_sty:"+name
                 +":"+doc.getLocation()));
            }
         }

         KeyValList userStyOpts = userSty.getOptions();
         KeyValList mainStyOpts = (mainSty == null ? null :  mainSty.getOptions());

         if (userStyOpts == null)
         {
            continue;
         }

         for (String opt : userStyOpts.keySet())
         {
            if (mainStyOpts == null || !mainStyOpts.containsKey(opt))
            {
               addErrorMessage(app.getLabelWithValues("error.book.mismatched_sty_opts",
               new String[]
               {
                  doc.getBaseName(),
                  name,
                  opt
               })
               +getInformationLink("mismatched_sty_opts:"+name+":"+opt
                + ":"+doc.getFile()));
            }
         }
      }
   }

   public void updateMainInfo()
      throws IOException,InterruptedException,InvalidBookException
   {
      preface = new Preface(this);

      File aux = getAuxFile();
      File mainFile = getFile();

      if (!aux.exists() || mainFile.lastModified() > aux.lastModified())
      {
         app.checkForInterrupt();
         int exitCode = latex();
         app.checkForInterrupt();

         if (exitCode != 0)
         {
            if (aux.exists())
            {
               if (!confirmContinue(
                   app.getLabelWithValue("error.exit.latex.cont", exitCode)))
               {
                  throw new CancelledException();
               }
            }
            else
            {
               String errMess = app.getLabelWithValues(
                     "error.exit.latex.failure",
                      new String[]
                      {
                         ""+exitCode,
                         app.getLaTeXApp(),
                         app.getProperties().getLaTeXOptions(),
                         getLocation().getAbsolutePath(),
                         getLogFile().getAbsolutePath()
                      }
                   );

               addErrorMessage(errMess);

               throw new InvalidBookException(errMess);
            }
         }
      }

      BufferedReader in = null;

      try
      {
         if (aux.exists())
         {
            importFiles.clear();
            hasSupplemental=false;
   
            app.setInfo("Parsing "+aux.getAbsolutePath());

            parseAux(aux);
         }
   
         File log = getLogFile();
   
         String line;
   
         if (log.exists())
         {
            app.setInfo(app.getLabelWithValue("message.parsing", log.getAbsolutePath()));
   
            in = new BufferedReader(new FileReader(log));
   
            while (logParse(in))
            {
               app.checkForInterrupt();
            }
   
            in.close();
            in = null;
         }
   
         in = new BufferedReader(new FileReader(mainFile));
   
         jmlrVolume = -1;
         bookVolume = -1;
         jmlrYear   = -1;
         workshopTitle = null;
         bookLocation  = null;
         bookDate      = null;
         bookfonts     = false;
   
         while ((line = in.readLine()) != null)
         {
            app.checkForInterrupt();
   
            for (int i = 0; i < bookPatterns.length; i++)
            {
               Matcher m = bookPatterns[i].matcher(line);

               if (m.matches())
               {
                  switch (i)
                  {
                     case BOOK_PATTERN_JMLRVOL:
                        try
                        {
                           jmlrVolume = Integer.parseInt(m.group(1));
                        }
                        catch (NumberFormatException e)
                        {
                        }
                     break;
                     case BOOK_PATTERN_VOL:
                        try
                        {
                           bookVolume = Integer.parseInt(m.group(1));
                        }
                        catch (NumberFormatException e)
                        {
                        }
                     break;
                     case BOOK_PATTERN_YEAR:
                        try
                        {
                           jmlrYear = Integer.parseInt(m.group(1));
                        }
                        catch (NumberFormatException e)
                        {
                        }
                     break;
                     case BOOK_PATTERN_WORKSHOP:
                        setWorkShop(m.group(1));
                     break;
                     case BOOK_PATTERN_BOOKFONTS:
                        bookfonts = true;
                     break;
                  }
               }
   
            }

         }
      }
      finally
      {
         if (in != null)
         {
            in.close();
         }
      }
   }

   protected void addImportFile(AuxData auxData, TeXParser parser, 
     boolean isSupplemental)
     throws IOException,InvalidBookException
   {
      String importLabel = auxData.getArg(0).toString(parser);

      if (getImportFile(importLabel) != null)
      {
         return;
      }

      String subdir = auxData.getArg(1).toString(parser);

      String basename = auxData.getArg(2).toString(parser);

      Path path = (new File(File.separatorChar == '/'
               ? subdir : subdir.replaceAll("/", File.separator))).toPath();

      String filename = basename;

      if (!(basename.endsWith(".tex") || basename.endsWith(".ltx")))
      {
         filename += ".tex";
      }

      Iterator<Path> it = path.iterator();
      FileElement currentDir = parent;

      while (it.hasNext())
      {
         String thisDir = it.next().toFile().getName();

         boolean found = false;

         for (BookFile child : currentDir.children)
         {
            if (child.getName().equals(thisDir))
            {
               if (!child.isDirectory())
               {
                  throw new IOException("Not a valid path: "+subdir);
               }

               currentDir = child;
               found = true;
               break;
            }
         }

         if (!found)
         {
            BookFile child = new BookFile(app, currentDir, thisDir, true);
            currentDir.children.add(child);

            currentDir = child;
         }
      }

      BookDocument importFile = null;

      // find filename in currentDir

      for (int i = 0; i < currentDir.getChildCount(); i++)
      {
         BookFile child = currentDir.children.get(i);

         if (child.getName().equals(filename))
         {
            if (child.isDirectory())
            {
               throw new IOException("File required, directory found: "
                    + (new File(path.toFile(), filename)));
            }

            importFile = child.toBookDocument();
         }
      }

      if (importFile == null)
      {
         importFile = new BookDocument(app, currentDir, basename);
         currentDir.addChild(importFile);
      }

      // does file exist?

      File file = importFile.getFile();

      if (!file.exists())
      {
         throw new InvalidBookException(app.getLabelWithValue(
              "error.importfile.missing", 
            file.getAbsolutePath()));
      }

      importFile.setLabel(importLabel);

      importFile.setSupplemental(isSupplemental);

      if (isSupplemental)
      {
         hasSupplemental = true;
      }

      importFiles.add(importFile);
   }

   public void parseAux(File auxFile)
    throws IOException
   {
      AuxParser auxParser = new AuxParser(new BookTeXApp(getApplication()))
      {
         protected void addPredefined()
         {
            super.addPredefined();

            addAuxCommand("jmlr@author", 2);
            addAuxCommand("jmlr@title", 2);
            addAuxCommand("jmlrbook@info", 2);
            addAuxCommand("jmlrbook@location", 1);
            addAuxCommand("jmlr@date", 1);
            addAuxCommand("jmlr@workshop", 1);
            addAuxCommand("@prefacestart", 2);
            addAuxCommand("@prefacefile", 2);
            addAuxCommand("@prefaceeditor", 1);
            addAuxCommand("@prefaceend", 1);
            addAuxCommand("@jmlr@import", 3);
            addAuxCommand("@jmlr@apdimport", 3);
            addAuxCommand("@new@articleauthor", 2);
         }
      };

      TeXParser parser = auxParser.parseAuxFile(auxFile);

      for (AuxData data : auxParser.getAuxData())
      {
         String name = data.getName();

         if (name.equals("bibdata"))
         {
            bibdata = data.getArg(0).toString(parser);
         }
         else if (name.equals("bibstyle"))
         {
            bibstyle = data.getArg(0).toString(parser);
         }
         else if (name.equals("jmlrbook@info"))
         {
            if (shortAuthor == null)
            {
               setShortAuthor(new TeXHtmlData(app, data.getArg(0), parser));
            }

            if (shortTitle == null)
            {
               setShortTitle(new TeXHtmlData(app, data.getArg(1), parser));
            }
         }
         else if (name.equals("jmlr@author"))
         {
            setAuthor(data, parser);
         }
         else if (name.equals("jmlr@title"))
         {
            setTitle(data, parser);
         }
         else if (name.equals("jmlrbook@location"))
         {
            setBookLocation(data, parser);
         }
         else if (name.equals("jmlr@date"))
         {
            setBookDate(data, parser);
         }
         else if (name.equals("jmlr@workshop"))
         {
            setWorkShop(data, parser);
         }
         else if (name.equals("@prefacestart"))
         {
            String startPage = data.getArg(0).toString(parser);

            try
            {
               setPrefaceStart(startPage,
                  Integer.parseInt(data.getArg(1).toString(parser)));
            }
            catch (NumberFormatException e)
            {
               setPrefaceStart(startPage, 0);
                 addErrorMessage(
                  app.getLabel("error.noparse.prefacenum"));
            }
         }
         else if (name.equals("@prefacefile"))
         {
            setPrefaceFileName(data.getArg(0).toString(parser), 
              data.getArg(1).toString(parser));
         }
         else if (name.equals("@prefaceeditor"))
         {
            addPrefaceEditor(new TeXHtmlData(app, data.getArg(0), parser));
         }
         else if (name.equals("@prefaceend"))
         {
            setPrefaceEnd(data.getArg(0).toString(parser));
         }
         else if (name.equals("@jmlr@import") || name.equals("@jmlr@apdimport"))
         {
            try
            {
               addImportFile(data, parser, name.equals("@jmlr@apdimport"));
            }
            catch (InvalidBookException e)
            {
               addErrorMessage(e.getMessage());
            }
            catch (IOException e)
            {
               addErrorMessage(e.getMessage());
               e.printStackTrace();
            }
         }
         else if (name.equals("@new@articleauthor"))
         {
            String label = data.getArg(0).toString(parser);
            String author = data.getArg(1).toString(parser);

            BookDocument importFile = null;

            for (int j = importFiles.size()-1; j >= 0; j--)
            {
               BookDocument currentIF = importFiles.get(j);

               if (currentIF.getLabel().equals(label))
               {
                  importFile = currentIF;
                  importFile.setAuthor(author);
                  break;
               }
            }

            if (importFile == null)
            {
               addErrorMessage(
                app.getLabelWithValues("error.importfile.not_found",
                 label, data.toString(parser)));
            }
         }
         else if (name.equals("newlabel"))
         {
            String label = data.getArg(0).toString(parser);

            boolean isStart;

            if (label.endsWith("jmlrstart"))
            {
               isStart = true;
            }
            else if (label.endsWith("jmlrend"))
            {
               isStart = false;
            }
            else
            {
               continue;
            }

            label = label.substring(0, label.lastIndexOf("jmlr"));

            if (label.isEmpty())
            {
               continue;
            }

            TeXObject arg = data.getArg(1);

            int pageNum = 0;

            String title = null;

            if (arg instanceof TeXObjectList 
             &&((TeXObjectList)arg).size() > 1)
            {
               TeXObject obj = ((TeXObjectList)arg).get(1);

               if (obj instanceof Group)
               {
                  obj = ((Group)obj).toList();
               }

               pageNum = Integer.parseInt(obj.toString(parser));

               if (isStart)
               {
                  obj = ((TeXObjectList)arg).get(2);

                  if (obj instanceof Group)
                  {
                     obj = ((Group)obj).toList();
                  }

                  title = obj.toString(parser);
               }
            }

            BookDocument importFile = null;

            for (int j = 0, n = importFiles.size(); j < n; j++)
            {
               BookDocument currentIF = importFiles.get(j);

               if (currentIF.getLabel().equals(label))
               {
                  importFile = currentIF;

                  if (isStart)
                  {
                     importFile.setStartPage(pageNum);
                     importFile.setTitle(title);
                  }
                  else
                  {
                     importFile.setEndPage(pageNum);
                  }

                  break;
               }
            }

            if (importFile == null)
            {
               addErrorMessage(
                app.getLabelWithValues("error.importfile.not_found",
                 label, data.toString(parser)));
            }
         }
      }
 
   }

   public BookDocument getImportFile(String label)
   {
      for (int i = 0, n = importFiles.size(); i < n; i++)
      {
         BookDocument doc = importFiles.get(i);

         if (doc.getLabel().equals(label))
         {
            return doc;
         }
      }

      return null;
   }

   public BookDocument getImportFileWithBaseName(String baseName)
   {
      for (int i = 0, n = importFiles.size(); i < n; i++)
      {
         BookDocument doc = importFiles.get(i);

         if (doc.getBaseName().equals(baseName))
         {
            return doc;
         }
      }

      return null;
   }

   public BookDocument getImportFile(int index)
   {
      return importFiles.get(index);
   }

   public void processImportFiles(boolean doRun)
      throws IOException,InterruptedException
   {
      String texArg = "";

      if (jmlrYear != -1)
      {
         texArg="\\expandafter\\def\\csname @jmlryear\\endcsname{"
               + jmlrYear+"}";
      }

      if (jmlrVolume != -1)
      {
         texArg += "\\expandafter\\def\\csname @jmlrvolume\\endcsname{"
                  + jmlrVolume+"}";
      }

      if (workshopTitle != null)
      {
         texArg += "\\expandafter\\def\\csname @jmlrworkshop\\endcsname{"
                 + getTeXWorkShop()+"}";
      }

      if (bookfonts)
      {
         texArg += "\\AtBeginDocument{\\InputIfFileExists{../bookfonts}{}{}\\normalfont }";
      }

      for (int i = 0, n = importFiles.size(); i < n; i++)
      {
         importFiles.get(i).process(false, doRun, true, texArg);
      }
   }

   public void checkConforms(boolean printVersion)
      throws IOException
   {
      if (!isJmlrbookClsAtLeast(2017, 3, 9))
      {
         addErrorMessage(app.getLabelWithValue("error.jmlrbookcls.tooold",
            "2017-03-09")
          + getInformationLink("tooold"));
      }

      if (jmlrbookClsVersion != null && jmlrClsVersion != null)
      {
         if (!jmlrbookClsVersion.equals(jmlrClsVersion))
         {
            addErrorMessage(app.getLabel("error.jmlrcls.book.mismatch")
             +getInformationLink("book.mismatch"));
         }
      }

      if (!app.isStrict() || isCiML())
      {
         return;
      }

      String prevTag = null;
      boolean alphabetical = true;
      String failedCompare = null;

      boolean failed=false;

      for (int i = 0, n = importFiles.size(); i < n; i++)
      {
         BookDocument importFile = importFiles.get(i);

         // Don't bother checking supplemental material

         if (importFile.isSupplemental())
         {
            continue;
         }

         String tag = importFile.getTag();

         if (tag == null)
         {
            app.warning(app.getLabelWithValue("warning.importfile.notag",
               importFile.getBaseName()));
            break;
         }

         Pattern p = Pattern.compile(tag+"\\d\\d[a-z]");

         Matcher m = p.matcher(importFile.getBaseName());

         boolean impfailed = false;

         if (!m.matches())
         {
            importFile.addErrorMessage(
               "<font class=nonconform>"
               +app.getLabelWithValue("error.jmlr.nonconform.filename",
                p.pattern())
               +"</font>"+getInformationLink("nonconform"));
            impfailed = true;
         }

         m = p.matcher(importFile.getLocation().getName());

         if (!m.matches())
         {
            importFile.addErrorMessage(
               "<font class=nonconform>"
                  +app.getLabel("error.jmlr.nonconform.dirname")
               +"</font>"+getInformationLink("nonconform"));
            impfailed = true;
         }

         if (impfailed)
         {
            importFile.addErrorMessage(
              "<font class=nonconform>"
              +app.getLabelWithValue("error.jmlr.seefaq", 
                   proceedingsFaq.toString())
              +"</font>"+getInformationLink("nonconform"));
            failed = impfailed;
         }

/*
         if (prevTag != null)
         {
            if (tag.compareTo(prevTag) < 0)
            {
               alphabetical = false;

               if (failedCompare == null)
               {
                  failedCompare = "'" + tag + "' &lt; '" + prevTag+"'";
               }
               else
               {
                  failedCompare += ", '" + tag + "' &lt; '" + prevTag+"'";
               }
            }
         }
*/

         prevTag = tag;
      }

      if (!failed)
      {
         failed = !alphabetical;
      }

      if (!alphabetical)
      {
         addErrorMessage(
            "<font class=nonconform>"
            +app.getLabelWithValue
               ("error.jmlr.nonconform.alpha", failedCompare)
            +"</font>"+getInformationLink("nonconform"));
      }

      if (preface == null || preface.getStart() == null)
      {
         addErrorMessage(
            "<font class=nonconform>"
            +app.getLabel("error.jmlr.nonconform.nopreface")
            +"</font>"+getInformationLink("nonconform"));
         failed = true;
      }
      else if (!preface.getStart().equals("1"))
      {
         preface.addErrorMessage(
            "<font class=nonconform>"
            +app.getLabelWithValue
            ("error.jmlr.nonconform.prefacestart", preface.getStart())
            +"</font>"+getInformationLink("nonconform"));
         failed = true;
      }

      if (failed)
      {
         addErrorMessage(
           "<font class=nonconform>"
           + app.getLabelWithValue("error.jmlr.seefaq", proceedingsFaq.toString())
           +"</font>"+getInformationLink("nonconform"));
      }
   }

   public void checkCls()
   {
      super.checkCls();

      for (int i = 0, n = importFiles.size(); i < n; i++)
      {
         BookDocument doc = importFiles.get(i);

         if (getJmlrClsLocation() != null && doc.getJmlrClsLocation() != null)
         {
            String comparator = null;

            switch (doc.compareJmlrCls(this))
            {
               case -1:
                  comparator = app.getLabel("error.comparator.older");
               break;
               case 1:
                  comparator = app.getLabel("error.comparator.newer");
               break;
            }

            if (comparator != null)
            {
               doc.addErrorMessage(app.getLabelWithValue
                  ("error.jmlrcls.mismatch", comparator)
                +getInformationLink("mismatch:"+doc.getLocation()));
            }
         }
      }
   }

   public Preface getPreface()
   {
      return preface;
   }

   public String getPrefaceSource()
   {
      return preface == null ? null : preface.getSource();
   }

   public String getPrefaceFileName()
   {
      if (preface == null) return null;

      String name = preface.getFileName();

      if (name == null) return null;

      if (name.endsWith(".tex") || name.endsWith(".ltx"))
      {
         return name;
      }

      return name+".tex";
   }

   public void setPrefaceFileName(String source, String name)
   {
      if (preface == null)
      {
         preface = new Preface(this);
      }

      preface.setFileName(source, name);
   }

   public String getPrefaceStart()
   {
      return preface == null ? null : preface.getStart();
   }

   public void setPrefaceStart(String representation, int number)
   {
      if (preface == null)
      {
         preface = new Preface(this);
      }

      preface.setStart(representation, number);
   }

   public int getPrefaceStartNumber()
   {
      return preface == null ? 0 : preface.getStartNumber();
   }

   public String getPrefaceEnd()
   {
      return preface == null ? null : preface.getEnd();
   }

   public void setPrefaceEnd(String representation)
   {
      if (preface == null)
      {
         preface = new Preface(this);
      }

      preface.setEnd(representation);
   }

   public Vector<TeXHtmlData> getPrefaceEditors()
   {
      return preface == null ? null : preface.getEditors();
   }

   public void addPrefaceEditor(TeXHtmlData name)
   {
      if (preface == null)
      {
         preface = new Preface(this);
      }

      preface.addEditor(name);
   }

   public File getPrefacePdfFile()
   {
      if (preface == null) return null;

      File file = preface.getDocFile();

      if (file == null) return null;

      String name = file.getName();

      return new File(file.getParentFile(), 
        name.substring(0, name.length()-3)+"pdf");
   }

   public Vector<BookDocument> getImportFiles()
   {
      return importFiles;
   }

   public String getAllErrorMessages()
   {
      String mess = getErrorMessages();

      String prefMess = getPrefaceErrorMessages();

      if (prefMess != null)
      {
         if (mess == null)
         {
            mess = prefMess;
         }
         else
         {
            mess += "\n" + prefMess;
         }
      }

      for (int i = 0, n = importFiles.size(); i < n; i++)
      {
         String importErrMess = importFiles.get(i).getErrorMessages();

         if (importErrMess != null)
         {
            if (mess == null)
            {
               mess = importErrMess;
            }
            else
            {
               mess += "\n" + importErrMess;
            }
         }
      }

      return mess;
   }


   public String getPrefaceErrorMessages()
   {
      return preface == null ? null : preface.getErrorMessages();
   }

   public int getBookVolume()
   {
      return bookVolume == -1 ? jmlrVolume : bookVolume;
   }

   public int getJmlrVolume()
   {
      return jmlrVolume;
   }

   public int getJmlrYear()
   {
      return jmlrYear;
   }

   public void setBookLocation(AuxData data, TeXParser parser)
   {
      bookLocation = new TeXHtmlData(app, data.getArg(0), parser);
   }

   public String getTeXLocation()
   {
      return bookLocation == null ? "" : bookLocation.getTeXString();
   }

   public void setBookDate(AuxData data, TeXParser parser)
   {
      bookDate = new TeXHtmlData(app, data.getArg(0), parser);
   }

   public String getHtmlDate()
   {
      return bookDate == null ? "" : bookDate.getHtml();
   }

   public String getTeXDate()
   {
      return bookDate == null ? "" : bookDate.getTeXString();
   }

   public void setWorkShop(String texSrc)
   {
      workshopTitle = new TeXHtmlData(app, texSrc);
   }

   public void setWorkShop(AuxData data, TeXParser parser)
   {
      workshopTitle = new TeXHtmlData(app, data.getArg(0), parser);
   }

   public String getHtmlWorkShop()
   {
      return workshopTitle == null ? "" : workshopTitle.getHtml();
   }

   public String getTeXWorkShop()
   {
      return workshopTitle == null ? "" : workshopTitle.getTeXString();
   }

   public boolean hasBookFontsFile()
   {
      return bookfonts;
   }

   public String getInfoContent()
   {
      StringBuffer buffer = new StringBuffer();

      buffer.append(String.format("<h1><a name=\"%s\">%s</a></h1><h2>%s</h2>",
        getBaseName(), getFile().getName(), getHtmlAbbrvTitle()));

      String workshop = getHtmlWorkShop();

      if (!workshop.isEmpty())
      {
         buffer.append(String.format("<h3>%s</h3>", workshop));
      }

      String location = getJmlrbookClsLocation();

      buffer.append(String.format(
        "<table><tr><th>%s</th><td>%s</td></tr><tr><th>%s</th><td>%s</td></tr><tr><th>%s</th><td>%s</td></tr>",
        app.getLabel("main.type"), 
        app.getLabel("main.type."+(isCiML() ? "ciml" : "jmlr")),
        app.getLabelWithValue("main.version", "jmlrbook.cls"),
        displayJmlrbookClsDetails(),
        app.getLabel("main.location"),
        (location==null?app.getLabel("error.unknown"):location)));

      location = getJmlrClsLocation();

      buffer.append(String.format(
        "<tr><th>%s</th><td>%s</td></tr><tr><th>%s</th><td>%s</td></tr>",
        app.getLabelWithValue("main.version", "jmlr.cls"),
        displayJmlrClsDetails(),
        app.getLabel("main.location"),
        (location==null?app.getLabel("error.unknown"):location)));

      if (jmlrVolume > -1)
      {
         buffer.append(String.format("<tr><th>%s</th><td>%d</td></tr>",
           app.getLabel("main.volume.jmlr"), jmlrVolume));
      }

      if (bookVolume > -1)
      {
         buffer.append(String.format("<tr><th>%s</th><td>%d</td></tr>",
           app.getLabel("main.volume.book"), bookVolume));
      }

      buffer.append("</table>");

      buffer.append(getInfoLinks());

      String src = getPrefaceSource();

      if (src != null)
      {
         String td = src;

         File file = preface.getSourceFile();
         File docFile = preface.getDocFile();

         if (file != null && file.exists())
         {
            td = String.format("<a href=\"%s\">%s</a>",
              file.toURI(), src);
         }

         if (!isCiML()
             && (file.equals(docFile) || file.equals(getFile())))
         {
            td += String.format(" <font class=error>%s</font>",
              app.getLabel("error.dependent_preface"));
         }

         buffer.append(String.format(
        "<h2><a name=preface>%s</a></h2><table><tr><th>%s</th><td>%s</td></tr>",
           app.getLabel("main.preface"),
           app.getLabel("main.source"),
           td));

         String prefFileName = (isCiML() ? null : getPrefaceFileName());

         if (prefFileName != null)
         {
            td = prefFileName;
            file = docFile;

            if (file != null && file.exists())
            {
               td = String.format("<a href=\"%s\">%s</a>",
                 file.toURI(), prefFileName);
            }

            if (file != null)
            {
               file = new File(file.getParentFile(), 
                 prefFileName.substring(0, prefFileName.lastIndexOf("."))+".pdf");

               if (file.exists())
               {
                  td += String.format(" (<a href=\"%s\">%s</a>)",
                    file.toURI(),
                    app.getLabel("main.view.pdf"));
               }
            }

            buffer.append(String.format("<tr><th>%s</th><td>%s</td></tr>",
              app.getLabel("main.file"), td));
         }

         String prefStart = getPrefaceStart();
         String prefEnd = getPrefaceEnd();

         if (prefStart.equals(prefEnd))
         {
            buffer.append(String.format("<tr><th>%s</th><td>%s</td></tr>",
               app.getLabel("main.page"),
               prefStart));
         }
         else
         {
            buffer.append(String.format("<tr><th>%s</th><td>%s-%s</td></tr>",
               app.getLabel("main.pages"),
               prefStart, prefEnd));
         }

         Vector<TeXHtmlData> editors = getPrefaceEditors();

         if (editors != null)
         {
            for (int i = 0, n = editors.size(); i < n; i++)
            {
               buffer.append(String.format(
                "<tr><th valign=\"top\">%s</th><td valign=\"top\">%s</td></tr>",
                app.getLabelWithValue("main.editor", (i+1)),
                editors.get(i).getHtml()));
            }
         }

         buffer.append("</table>");
      }

      String errMess = getPrefaceErrorMessages();

      if (errMess != null)
      {
         errMess = errMess.replaceAll("\n", "<br>");

         buffer.append(String.format("<font class=error>%s</font><br>",
           errMess));
      }

      if (importFiles != null)
      {
         for (int i = 0, n = importFiles.size(); i < n; i++)
         {
            buffer.append(importFiles.get(i).getInfoContent());
         }
      }

      return buffer.toString();
   }

   public String getPdfLink()
   {
      StringBuffer buffer = new StringBuffer();

      if (isCiML())
      {
         File pdfFile = getPdfFile();

         if (pdfFile.exists())
         {
            buffer.append(String.format(
              "<tr><td><a href=\"%s\">%s</a></td><td>%s</td></tr>",
               pdfFile.toURI(),
               app.getLabel("main.view.online"),
               app.getLabelWithValue("main.last_modified", 
                    dateFormat.format(new Date(pdfFile.lastModified())))));

         }

         pdfFile = getPdfFile(true);

         if (pdfFile.exists())
         {
            buffer.append(String.format(
               "<tr><td><a href=\"%s\">%s</a></td><td>%s</td></tr>",
               pdfFile.toURI(),
               app.getLabel("main.view.print"),
               app.getLabelWithValue("main.last_modified", 
                    dateFormat.format(new Date(pdfFile.lastModified())))));
         }
      }
      else
      {
         buffer.append(super.getPdfLink());
      }

      return buffer.toString();
   }

   public BookNode getBookNode()
   {
      return bookNode;
   }

   public boolean isCiML()
   {
      if (clsFile == null) return false;

      KeyValList options = clsFile.getOptions();

      TeXObject val = options.get("7x10");

      return (val != null);
   }

   public boolean isPMLR()
   {
      if (clsFile == null) return false;

      KeyValList options = clsFile.getOptions();

      TeXObject val = options.get("pmlr");

      return (val != null);
   }

   public boolean isWCP()
   {
      if (clsFile == null) return false;

      KeyValList options = clsFile.getOptions();

      TeXObject val = options.get("wcp");

      return (val != null);
   }

   public void setLogo(TeXObject logoObj)
     throws IOException
   {
      if (logoObj == null)
      {
         logo = null;
         return;
      }

      if (logoObj instanceof TeXObjectList)
      {
         TeXObjectList list = (TeXObjectList)logoObj;

         while (list.size() > 0)
         {
            TeXObject obj = list.popStack(null);

            if (obj instanceof ControlSequence
            && ((ControlSequence)obj).getName().equals("includegraphics"))
            {
               TeXObject opt = list.popArg(null, '[', ']');

               logoObj = list.popArg(null);
               break;
            }
         }
      }

      if (logoObj instanceof TeXObjectList)
      {
         TeXObjectList list = (TeXObjectList)logoObj;

         logo = "";

         for (int i = 0; i < list.size(); i++)
         {
            TeXObject obj = list.get(i);

            logo += obj.toString();
         }
      }
      else
      {
         logo = logoObj.toString();
      }
   }

   public void setLogoUrl(TeXObject logoUrlObj)
   {
      if (logoUrlObj == null)
      {
         logoUrl = null;
         return;
      }

      if (logoUrlObj instanceof TeXObjectList)
      {
         TeXObjectList list = (TeXObjectList)logoUrlObj;

         logoUrl = "";

         for (int i = 0; i < list.size(); i++)
         {
            TeXObject obj = list.get(i);

            logoUrl += obj.toString();
         }
      }
      else
      {
         logoUrl = logoUrlObj.toString();
      }
   }

   public String getLogoUrl()
   {
      return logoUrl;
   }

   public String getLogo()
   {
      return logo;
   }

   public void writeBibFieldStart(String entryTag, PrintWriter out)
     throws IOException
   {
      out.println(String.format("@Proceedings{%s,", entryTag));

      String shortTitle = getTeXShortTitle().replaceFirst("'?\\d+$", "").trim();

      out.println(String.format("  shortname = {%s},", shortTitle));
      out.println(String.format("  year = %d,", jmlrYear));
      out.println(String.format("  volume = %d,", jmlrVolume));
   }

   public String getBibTitleField()
   {
      return "booktitle";
   }

   public String getBibAuthorField()
   {
      return "editor";
   }

   public void createBib(File bibFile) throws IOException
   {
      PrintWriter out = null;
      app.debug("Creating "+bibFile);

      try
      {
         out = new PrintWriter(new FileWriter(bibFile));

         String proctag = getTeXShortTitle().replaceAll("'", "").trim().toLowerCase();

         String bibAuthors = writeBibEntry(proctag, out);

         out.println();

         if (preface != null)
         {
            preface.writeBibEntry(proctag, out, bibAuthors);
            out.println();
         }

         for (BookDocument doc : importFiles)
         {
            String entryTag = doc.getBaseName();

            doc.writeBibEntry(entryTag, proctag, out);
            out.println();
         }
      }
      finally
      {
         if (out != null)
         {
            out.close();
         }
      }
   }

   public String writeBibEntry(String entryTag, PrintWriter out)
     throws IOException
   {
      writeBibFieldStart(entryTag, out);

      out.println(String.format("  %s = {%s},",
        getBibTitleField(),
        getTeXTitle()));

      PreambleParser listener 
        = new PreambleParser(new BookTeXApp(getApplication()));

      TeXParser parser = new TeXParser(listener);
      String bibAuthors = getBibAuthor(parser);

      out.println(String.format("  %s = {%s}",
         getBibAuthorField(),
         bibAuthors));

      out.println("}");

      return bibAuthors;
   }

   public String writeBibEntry(String entryTag, String crossref,
     PrintWriter out)
     throws IOException
   {
      return writeBibEntry(entryTag, out);
   }

   public void compress(File zipFile) throws IOException
   {
      app.debug("Creating "+zipFile);

      BufferedInputStream origin = null;
      ZipOutputStream out = null;

      try
      {
         FileOutputStream dest = new FileOutputStream(zipFile);
         out = new ZipOutputStream(new BufferedOutputStream(dest));
         out.setMethod(ZipOutputStream.DEFLATED);

         File[] bibFiles = getLocation().listFiles(new java.io.FileFilter()
         {
            public boolean accept(File file)
            {
               return !file.isDirectory() && file.getName().toLowerCase().endsWith(".bib");
            }
         });

         byte data[] = new byte[BUFFER];

         if (bibFiles == null || bibFiles.length == 0)
         {
            File bibFile = new File(getLocation(), 
             getHtmlShortTitle().replaceAll("[^\\w]+", "")+".bib");
            createBib(bibFile);

            bibFiles = new File[1];
            bibFiles[0] = bibFile;
         }

         for (File bibFile : bibFiles)
         {
            app.setInfo(bibFile.toString());
            app.debug("Adding "+bibFile);
            FileInputStream fi = new FileInputStream(bibFile);
            origin = new BufferedInputStream(fi, BUFFER);

            ZipEntry entry = new ZipEntry(bibFile.getName());
            out.putNextEntry(entry);

            int count;
   
            while ((count = origin.read(data, 0, BUFFER)) != -1)
            {
               out.write(data, 0, count);
            }

            origin.close();
            origin = null;
         }

         File pdfFile = getPrefacePdfFile();
         int count;
   
         if (pdfFile != null && pdfFile.exists())
         {
            app.setInfo(pdfFile.toString());
            app.debug("Adding "+pdfFile);
            FileInputStream fi = new FileInputStream(pdfFile);
            origin = new BufferedInputStream(fi, BUFFER);

            ZipEntry entry = new ZipEntry(pdfFile.getName());
            out.putNextEntry(entry);

            while ((count = origin.read(data, 0, BUFFER)) != -1)
            {
               out.write(data, 0, count);
            }

            origin.close();
            origin = null;
         }

         for (BookDocument doc : importFiles)
         {
            pdfFile = doc.getPdfFile();

            if (pdfFile.exists())
            {
               app.setInfo(pdfFile.toString());
               app.debug("Adding "+pdfFile);
               FileInputStream fi = new FileInputStream(pdfFile);
               origin = new BufferedInputStream(fi, BUFFER);

               ZipEntry entry = new ZipEntry(pdfFile.getName());
               out.putNextEntry(entry);

               while ((count = origin.read(data, 0, BUFFER)) != -1)
               {
                  out.write(data, 0, count);
               }

               origin.close();
               origin = null;
            }

            File[] suppFiles = doc.getSupplementalFiles();

            if (suppFiles != null)
            {
               for (File suppFile : suppFiles)
               {
                  app.setInfo(suppFile.toString());
                  app.debug("Adding "+suppFile);
                  FileInputStream fi = new FileInputStream(suppFile);
                  origin = new BufferedInputStream(fi, BUFFER);

                  ZipEntry entry = new ZipEntry(suppFile.getName());
                  out.putNextEntry(entry);

                  while ((count = origin.read(data, 0, BUFFER)) != -1)
                  {
                     out.write(data, 0, count);
                  }

                  origin.close();
                  origin = null;
               }
            }
         }

      }
      finally
      {
         if (origin != null)
         {
            origin.close();
         }

         if (out != null)
         {
            out.close();
         }
      }
   }

   private static final int BUFFER = 2048;

   private Vector<BookDocument> importFiles;

   private int bookVolume = -1, jmlrVolume = -1, jmlrYear = -1;
   private File htmlIndexFile;

   private boolean bookfonts;

   private BookNode bookNode;

   private Preface preface;

   private TeXHtmlData workshopTitle, bookLocation, bookDate;

   private String logo, logoUrl;

   private boolean hasSupplemental;

   private static final Pattern[] htmlPatterns = new Pattern[]
   {
      Pattern.compile("[^%]*\\\\documentclass *(?:\\[([^\\]]*)\\])?.*"),
      Pattern.compile("[^%]*\\\\begin *\\{document\\}.*"),
      Pattern.compile("[^%]*\\\\end *\\{document\\}.*"),
      Pattern.compile("[^%]*\\\\begin *\\{preface\\}.*"),
      Pattern.compile("[^%]*\\\\end *\\{preface\\}.*"),
      Pattern.compile("[^%]*\\\\begin *\\{abstract\\}.*"),
      Pattern.compile("[^%]*\\\\end *\\{abstract\\}.*"),
      Pattern.compile("[^%]*\\\\logo\\s*(\\[[^\\]]*\\])?\\s*\\{\\\\includegraphics\\s*(?:\\[[^\\]]*\\])?\\{([^\\}]*)\\}\\}.*"),
      Pattern.compile("[^%]*\\\\maketitle.*"),
   };

   private static final int HTML_PATTERN_DOCCLS=0;
   private static final int HTML_PATTERN_BEGINDOC=1;
   private static final int HTML_PATTERN_ENDDOC=2;
   private static final int HTML_PATTERN_BEGINPREF=3;
   private static final int HTML_PATTERN_ENDPREF=4;
   private static final int HTML_PATTERN_BEGINABS=5;
   private static final int HTML_PATTERN_ENDABS=6;
   private static final int HTML_PATTERN_LOGO=7;
   private static final int HTML_PATTERN_MAKETITLE=8;

   private static final Pattern[] bookPatterns = new Pattern[]
   {
      Pattern.compile("[^%]*\\\\jmlrvolume\\{(\\d+)\\}.*"),
      Pattern.compile("[^%]*\\\\volume\\{(\\d+)\\}.*"),
      Pattern.compile("[^%]*\\\\jmlryear\\{(\\d+)\\}.*"),
      Pattern.compile("[^%]*\\\\jmlrworkshop\\{([^\\}]+)\\}.*"),
      Pattern.compile("[^%]*\\\\input\\{bookfonts\\}.*"),
   };

   private static final int BOOK_PATTERN_JMLRVOL=0;
   private static final int BOOK_PATTERN_VOL=1;
   private static final int BOOK_PATTERN_YEAR=2;
   private static final int BOOK_PATTERN_WORKSHOP=3;
   private static final int BOOK_PATTERN_BOOKFONTS=4;
   
   private static URI proceedingsFaq;

}

class Preface implements TreeNode
{
    private Preface()
    {
    }

    public Preface(Book book)
    {
       this.book = book;
       label = "preface";
       editors = new Vector<TeXHtmlData>();
    }
   
    public void addEditor(TeXHtmlData editor)
    {
       if (!editors.contains(editor))
       {
          editors.add(editor);
       }
    }
   
    public void setStart(String representation, int number)
    {
       start = representation;
       startNumber = number;
    }
   
    public void setEnd(String representation)
    {
       end = representation;
    }
   
    public void setFileName(String source, String name)
    {
       this.source = source;
       this.name = name.contains(".") ? name : name+".tex";
    }
   
    public String getSource()
    {
       return source;
    }

    public File getSourceFile()
    {
       if (source == null) return null;

       String[] split = source.split("/");

       File parent = book.getLocation();
       File srcFile = null;

       for (int i = 0; i < split.length; i++)
       {
          srcFile = new File(parent, split[i]);
          parent = srcFile;
       }

       return srcFile;
    }

    public File getDocFile()
    {
       if (name == null)
       {
          return null;
       }

       File parent = book.getLocation();

       File file = new File(parent, name);

       if (file.exists())
       {
          return file;
       }

       String[] split = source.split("/");

       for (int i = 0, n = split.length-1; i < n; i++)
       {
          parent = new File(parent, split[i]);
          file = new File(parent, name);

          if (file.exists()) return file;
       }

       return null;
    }

    public void latex() throws IOException,InterruptedException
    {
       File file = getDocFile();

       if (file == null || !file.exists())
       {
          throw new FileNotFoundException(MakeJmlrBookGUI.getLabelWithValue(
            "error.file_not_found",
             file == null ? name : file.toString()));
       }

       File dir = file.getParentFile();

       String latexOptions = book.getApplication().getProperties().getLaTeXOptions();

       String[] opts = latexOptions.split(" +");
       String[] cmd = new String[opts.length+4];

       int idx = 0;

       cmd[idx++] = book.getApplication().getLaTeXApp();
       cmd[idx++] = "-interaction";
       cmd[idx++] = "batchmode";

       for (int i = 0; i < opts.length; i++)
       {
          cmd[idx++] = opts[i];
       }

       cmd[idx] = name;

       int exitCode = book.getApplication().execCommandAndWaitFor(cmd, null, dir);

       book.getApplication().debug("Exit code: "+exitCode);

       if (exitCode != 0)
       {
          StringBuffer buff = new StringBuffer();

          for (int i = 1; i < cmd.length; i++)
          {
             buff.append(String.format("\"%s\" ", cmd[i]));
          }

         String errMess = 
            book.getApplication().getLabelWithValues("error.exit.latex.failure",
               new String[]
               {
                  ""+exitCode,
                  cmd[0],
                  buff.toString(),
                  dir.getAbsolutePath(),
                  name+".log"
               });

         addErrorMessage(errMess);

         throw new IOException(errMess);
       }
    }

    public void clearAux()
    {
       File texFile = getDocFile();

       if (texFile == null) return;

       File parent = texFile.getParentFile();

       String baseName = name.substring(0, name.lastIndexOf("."));

       book.removeAuxFile(parent, name+".aux");
       book.removeAuxFile(parent, name+".out");
    }

    public void writeBibEntry(String crossref, PrintWriter out, String authors)
      throws IOException
    {
       String tag = name.substring(0, name.length()-4);

       out.println("@InProceedings{"+tag+",");

       out.println("  section = {preface},");
       out.println("  title = {Preface},");
       out.println(String.format("  authors = {%s},", authors));

       out.println(String.format("  pages = {%s-%s},", getStart(), getEnd()));
       out.println(String.format("  crossref={%s},", crossref));
       out.println(String.format("  abstract = {Preface to %s %d}",
         book.getTeXShortTitle(), book.getJmlrYear()));

       out.println("}");
    }

    public String getFileName()
    {
       return name;
    }

    public String getStart()
    {
       return start;
    }

    public int getStartNumber()
    {
       return startNumber;
    }

    public String getEnd()
    {
       return end;
    }

    public String getPages()
    {
       String pp = start;

       if (!start.equals(end))
       {
          pp += "--"+end;
       }

       return pp;
    }
   
    public Vector<TeXHtmlData> getEditors()
    {
       return editors;
    }

    public void addErrorMessage(String message)
    {
       if (errorMessages == null)
       {
          errorMessages = message;
       }
       else if (errorMessages.indexOf(message) == -1)
       {
          errorMessages += "\n" + message;
       }
    }

    public String getErrorMessages()
    {
       return errorMessages;
    }

    public void clearErrorMessages()
    {
       errorMessages = null;
    }

    public Enumeration children()
    {
       return null;
    }

    public boolean getAllowsChildren()
    {
       return false;
    }

    public TreeNode getChildAt(int childIndex)
    {
       return null;
    }

    public int getChildCount()
    {
       return 0;
    }

    public int getIndex(TreeNode node)
    {
       return -1;
    }

    public TreeNode getParent()
    {
       return book.getBookNode();
    }

    public boolean isLeaf()
    {
       return true;
    }

    public String toString()
    {
       return label;
    }
   
    private String start, end, source, name, label;
    public int startNumber;
    private Vector<TeXHtmlData> editors;

    private String errorMessages;

    private Book book;
}

class BookNode implements TreeNode
{
   public BookNode(Book book)
   {
      this.book = book;
   }

   public Enumeration children()
   {
      return new BookImportEnumeration(book);
   }

   public boolean getAllowsChildren()
   {
      return true;
   }

   public TreeNode getChildAt(int childIndex)
   {
      Preface preface = book.getPreface();

      if (preface == null)
      {
         return book.getImportFiles().get(childIndex).getImportNode(book);
      }
      else
      {
         if (childIndex == 0)
         {
            return preface;
         }
         else
         {
            return book.getImportFiles().get(childIndex-1).getImportNode(book);
         }
      }
   }

   public int getChildCount()
   {
      Preface preface = book.getPreface();

      return (preface == null ? 0 : 1)+book.getImportFiles().size();
   }

   public int getIndex(TreeNode node)
   {
      if (node == null)
      {
         return -1;
      }

      Preface preface = book.getPreface();

      if (node == preface)
      {
         return 0;
      }

      int idx = preface == null ? 0 : 1;

      for (int i = 0, n = book.getImportFiles().size(); i < n; i++)
      {
         BookDocument doc = book.getImportFiles().get(i);

         if (node.equals(doc.getImportNode()))
         {
            return idx+i;
         }
      }

      return -1;
   }

   public TreeNode getParent()
   {
      return null;
   }

   public boolean isLeaf()
   {
      return getChildCount() == 0;
   }

   public String toString()
   {
      return book.getBaseName();
   }

   private Book book;
}

class BookImportEnumeration implements Enumeration<TreeNode>
{
   public BookImportEnumeration(Book book)
   {
      this.book = book;
      currentIdx = 0;
   }

   public int size()
   {
      return (book.getPreface() == null ? 0 : 1) + book.getImportFiles().size();
   }

   public boolean hasMoreElements()
   {
      return currentIdx < size()-1;
   }

   public TreeNode nextElement()
     throws NoSuchElementException
   {
      TreeNode elem = null;

      Preface preface = book.getPreface();

      if (preface == null)
      {
         try
         {
            elem = book.getImportFiles().get(currentIdx).getImportNode(book);
         }
         catch (ArrayIndexOutOfBoundsException e)
         {
            throw new NoSuchElementException();
         }
      }
      else
      {
         if (currentIdx == 0)
         {
            elem = preface;
         }
         else
         {
            try
            {
               elem = book.getImportFiles().get(currentIdx-1).getImportNode(book);
            }
            catch (ArrayIndexOutOfBoundsException e)
            {
               throw new NoSuchElementException();
            }
         }
      }

      currentIdx++;

      return elem;
   }

   private Book book;
   private int currentIdx;
}

class TeXModWalkListener implements BookWalkListener
{
   public TeXModWalkListener(long pdfMod)
   {
      this.pdfMod = pdfMod;
   }

   public FileElement walkAction(FileElement element)
   {
      if (element.getName().toLowerCase().endsWith(".tex"))
      {
         File file = element.getFile();

         long mod = file.lastModified() ;

         if (mod > pdfMod)
         {
            runRequired = true;

            if (MakeJmlrBookGUI.inDebugMode())
            {
               MakeJmlrBookGUI.printAndLogMessage("Run required: tex file modified since pdf mod."
                 +"\n  tex: "+file.getAbsolutePath()
                 +"\n  tex mod: "+(new Date(mod))
                 +"\n  pdf mod: "+(new Date(pdfMod)));
            }

            return element;
         }

      }

      return null;
   }

   public boolean isRunRequired()
   {
      return runRequired;
   }

   private boolean runRequired = false;
   private long pdfMod;
}
