/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.io.*;
import java.util.*;
import java.awt.event.ActionListener;

import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JToolBar;

public class JMLRBookProperties extends Properties
{
   public JMLRBookProperties(File propFile, File recentFile,
      File loggingFile, boolean loadFile, MakeJmlrBookGUI application)
     throws IOException
   {
      super();

      app = application;

      this.propFile = propFile;
      this.recentFile = recentFile;
      this.loggingFile = loggingFile;

      recentList = new Vector<String>();

      if (propFile != null && propFile.exists() && loadFile)
      {
         // if propFile exists, load it

         BufferedReader reader = 
            new BufferedReader(new FileReader(propFile));

         load(reader);

         reader.close();
      }
      else
      {
         // set default values

         setDefaults();
      }

      if (recentFile != null && recentFile.exists())
      {
         BufferedReader in = new BufferedReader(new FileReader(recentFile));

         loadRecentFiles(in);

         in.close();
      }
   }

   public JMLRBookProperties(MakeJmlrBookGUI application)
   {
      super();

      app = application;

      propFile       = null;
      recentFile     = null;

      setDefaults();
   }

   protected void setDefaults()
   {
      setProperty("directory", ".");
      setProperty("latexopts", "");
      setProperty("bibtexopts", "");
      setProperty("strict", "true");
      setProperty("toolbarorient", ""+JToolBar.HORIZONTAL);
   }

   public static void initResourceDir()
   {
      String home = System.getProperty("user.home");

      if (home != null)
      {
         // first check if directory ~/.makejmlrbook exists

         resourceDir = new File(home, ".makejmlrbook");

         if (!resourceDir.exists())
         {
            resourceDir.mkdir();
         }
         else if (!resourceDir.isDirectory())
         {
            // check if directory ~/makejmlrbook-settings exists

            resourceDir = new File(home, "makejmlrbook-settings");

            if (!resourceDir.isDirectory())
            {
               resourceDir = null;
            }
         }
      }
   }

   public static JMLRBookProperties fetchProperties(MakeJmlrBookGUI application)
     throws IOException
   {
      File settings   = null;
      File recent     = null;
      File logging    = null;

      if (resourceDir == null)
      {
         initResourceDir();
      }

      boolean loadFile = false;

      if (resourceDir != null)
      {
         settings = new File(resourceDir, baseName);
         recent = new File(resourceDir, recentName);
         logging = new File(resourceDir, loggingName);
         loadFile = true;
      }

      return new JMLRBookProperties(settings, recent, logging, loadFile, application);
   }

   public void save()
      throws IOException
   {
      if (propFile == null)
      {
         String home = System.getProperty("user.home");

         if (home == null)
         {
            throw new IOException(app.getLabel("error.home_null"));
         }

         File parent;

         if (System.getProperty("os.name").startsWith("Win"))
         {
            parent = new File(home, "makejmlrbook-settings");
         }
         else
         {
            parent = new File(home, ".makejmlrbook");
         }

         if (!parent.mkdir())
         {
            throw new IOException(app.getLabelWithValue("error.io.cant_mkdir",
               parent.getAbsolutePath()));
         }

         propFile = new File(parent, baseName);
      }

      String setting = getDefaultDirectorySetting();

      if (setting == null)
      {
         setting = "home";
      }

      if (setting.equals("home"))
      {
         setDefaultDirectory(System.getProperty("user.home"));
      }
      else if (setting.equals("last"))
      {
         setDefaultDirectory(app.getCurrentDirectory());
      }

      PrintWriter out = new PrintWriter(new FileWriter(propFile));

      store(out, "makejmlrbookgui properties");

      out.close();

      if (recentFile == null)
      {
         recentFile = new File(propFile.getParentFile(), recentName);
      }

      out = new PrintWriter(new FileWriter(recentFile));

      for (int i = 0, n = recentList.size(); i < n; i++)
      {
         out.println(recentList.get(i));
      }

      out.close();
   }

   private void loadRecentFiles(BufferedReader in)
     throws IOException
   {
      String line;

      while ((line = in.readLine()) != null)
      {
         recentList.add(line);
      }
   }

   public void addRecentFile(String fileName)
   {
      recentList.remove(fileName); // just in case it's already in the list
      recentList.add(fileName);
   }

   public void setRecentFiles(JMenu menu, ActionListener listener)
   {
      if (recentList == null) return;

      if (menu.getMenuComponentCount() > 0)
      {
         menu.removeAll();
      }

      int lastIdx = recentList.size()-1;

      int n = Math.min(MAX_RECENT_FILES-1, lastIdx);

      for (int i = 0; i <= n; i++)
      {
         File file = new File(recentList.get(lastIdx-i));
         String num = ""+i;
         JMenuItem item = new JMenuItem(num+": "+file.getName());
         item.setMnemonic(num.charAt(0));
         item.setToolTipText(file.getAbsolutePath());
         item.setActionCommand(num);
         item.addActionListener(listener);

         menu.add(item);
      }
   }

   public String getRecentFileName(int i)
   {
      return recentList.get(recentList.size()-1-i);
   }

   public void clearRecentList()
   {
      recentList.clear();
   }

   public String getDefaultDirectory()
   {
      return getProperty("directory");
   }

   public void setDefaultDirectory(String dir)
   {
      if (dir != null)
      {
         setProperty("directory", dir);
      }
   }

   public String getDefaultDirectorySetting()
   {
      return getProperty("directory.setting");
   }

   public void setDefaultDirectorySetting(String setting)
   {
      if (setting != null)
      {
         setProperty("directory.setting", setting);
      }
   }

   public void setDefaultHomeDir()
   {
      setDefaultDirectorySetting("home");
   }

   public void setDefaultLastDir()
   {
      setDefaultDirectorySetting("last");
   }

   public void setDefaultCustomDir(String dir)
   {
      if (dir == null)
      {
         setDefaultDirectorySetting("last");
      }
      else
      {
         setDefaultDirectorySetting("custom");
         setDefaultDirectory(dir);
      }
   }

   public String getLaTeXOptions()
   {
      String defOpt = "-interaction=batchmode";
      String opts = getProperty("latexopts");

      if (opts == null)
      {
         opts = defOpt;
      }
      else
      {
         opts = defOpt + " " + opts;
      }

      if (useFileLine())
      {
         opts += " -file-line-error";
      }

      return opts;
   }


   public String getLaTeXApp()
   {
      return getProperty("latexapp");
   }

   public void setLaTeXApp(String appPath)
   {
      if (appPath == null)
      {
         setProperty("latexapp", "");
      }
      else
      {
         setProperty("latexapp", appPath);
      }
   }

   public String getBibTeXOptions()
   {
      String defOpt = "-terse";
      String opts = getProperty("bibtexopts");

      if (opts == null)
      {
         opts = defOpt;
      }
      else
      {
         opts = defOpt + " " + opts;
      }

      return opts;
   }

   public String getBibTeXApp()
   {
      return getProperty("bibtexapp");
   }

   public void setBibTeXApp(String appPath)
   {
      if (appPath == null)
      {
         setProperty("bibtexapp", "");
      }
      else
      {
         setProperty("bibtexapp", appPath);
      }
   }

   public String getKpsewhichApp()
   {
      return getProperty("kpsewhichapp");
   }

   public void setKpsewhichApp(String appPath)
   {
      if (appPath == null)
      {
         setProperty("kpsewhichapp", "");
      }
      else
      {
         setProperty("kpsewhichapp", appPath);
      }
   }

   public String getEpstopdfApp()
   {
      return getProperty("epstopdfapp");
   }

   public void setEpstopdfApp(String appPath)
   {
      if (appPath == null)
      {
         setProperty("epstopdfapp", "");
      }
      else
      {
         setProperty("epstopdfapp", appPath);
      }
   }

   public String getWmftoepsApp()
   {
      return getProperty("wmftoepsapp");
   }

   public void setWmftoepsApp(String appPath)
   {
      if (appPath == null)
      {
         setProperty("wmftoepsapp", "");
      }
      else
      {
         setProperty("wmftoepsapp", appPath);
      }
   }

   public String getPdfViewer()
   {
      return getProperty("pdfviewer");
   }

   public void setPdfViewer(String path)
   {
      if (path == null)
      {
         setProperty("pdfviewer", "");
      }
      else
      {
         setProperty("pdfviewer", path);
      }
   }

   public String getTextViewer()
   {
      return getProperty("textviewer");
   }

   public void setTextViewer(String path)
   {
      setProperty("textviewer", path);
   }

   public String getHtmlViewer()
   {
      return getProperty("htmlviewer");
   }

   public void setHtmlViewer(String path)
   {
      if (path == null)
      {
         setProperty("htmlviewer", "");
      }
      else
      {
         setProperty("htmlviewer", path);
      }
   }

   public String getImageViewer()
   {
      return getProperty("imageviewer");
   }

   public void setImageViewer(String path)
   {
      if (path == null)
      {
         setProperty("imageviewer", "");
      }
      else
      {
         setProperty("imageviewer", path);
      }
   }

   public boolean isBeepEnabled()
   {
      String val = getProperty("beep", "false");

      return val.equals("true");
   }

   public void setBeepEnabled(boolean isEnabled)
   {
      setProperty("beep", isEnabled ? "true" : "false");
   }

   public boolean showDetailsIcon()
   {
      String val = getProperty("details_icon", "true"); 

      return val.equals("true");
   }

   public boolean showDetailsText()
   {
      String val = getProperty("details_text", "false"); 

      return val.equals("true");
   }

   public void setShowDetailsIcon(boolean state)
   {
      setProperty("details_icon", state ? "true" : "false");
   }

   public void setShowDetailsText(boolean state)
   {
      setProperty("details_text", state ? "true" : "false");
   }

   public boolean isStrict()
   {
      String val = getProperty("strict", "true"); 

      return val.equals("true");
   } 

   public void setStrict(boolean isStrict)
   {
      setProperty("strict", isStrict ? "true" : "false");
   }

   public boolean useFileLine()
   {
      String val = getProperty("file-line", "false"); 

      return val.equals("true");
   }

   public void setUseFileLine(boolean enable)
   {
      setProperty("file-line", enable ? "true" : "false");
   }

   public int getToolBarOrientation()
   {
       String val = getProperty("toolbarorient");

       if (val == null)
       {
          return JToolBar.HORIZONTAL;
       }

       try
       {
          return Integer.parseInt(val);
       }
       catch (NumberFormatException e)
       {
       }

       return JToolBar.HORIZONTAL;
   }

   public void setToolBarOrientation(int orient)
   {
      setProperty("toolbarorient", ""+orient);
   }

   public String getToolBarPosition()
   {
      String val = getProperty("toolbarpos");

      return (val == null ? "North" : val);
   }

   public void setToolBarPosition(String pos)
   {
      if (pos == null)
      {
         setProperty("toolbarpos", "North");
      }
      else
      {
         setProperty("toolbarpos", pos);
      }
   }

   public void setLogoUrl(String url)
   {
      if (url == null)
      {
         setProperty("logourl", "");
      }
      else
      {
         setProperty("logourl", url);
      }
   }

   public String getLogoUrl()
   {
      return getProperty("logourl");
   }

   public String getICCFile()
   {
      return getProperty("iccfile");
   }

   public void setICCFile(String iccfile)
   {
      setProperty("iccfile", iccfile);
   }

   public String getGSApp()
   {
      return getProperty("gs");
   }

   public void setGSApp(String gspath)
   {
      setProperty("gs", gspath);
   } 

   public String getTeXHash()
   {
      return getProperty("texhash", "mktexlsr");
   }

   public void setTeXHash(String texhash)
   {
      setProperty("texhash", texhash);
   }

   public long getMaxProcessTime()
   {
      String val = getProperty("max_process_time");

      long maxTime = 1000*60*5;

      if (val != null)
      {
         try
         {
            maxTime = Long.parseLong(val);
         }
         catch (NumberFormatException e)
         {
         }
      }

      return maxTime;
   }

   public void setMaxProcessTime(long maxProcessTime)
   {
      setProperty("max_process_time", ""+maxProcessTime);
   }

   public void setVersion(String version)
   {
      setProperty("version", version);
   }

   public String getVersion()
   {
      return getProperty("version");
   }

   public boolean isLogging()
   {
      String val = getProperty("logging");

      boolean isOn = false;

      if (val != null)
      {
         isOn = Boolean.parseBoolean(val);
      }

      return isOn;
   }

   public void enableLogging(boolean isOn)
   {
      setProperty("logging", ""+isOn);
   }

   public File getLoggingFile()
   {
      if (loggingFile == null)
      {
         if (resourceDir == null)
         {
            initResourceDir();
         }

         if (resourceDir != null)
         {
            loggingFile = new File(resourceDir, loggingName);
         }
      }

      return loggingFile;
   }

   public MakeJmlrBookGUI getApplication()
   {
      return app;
   }

   private File propFile;

   private static final String baseName = "makejmlrbook.prop";
   private static String recentName = "recentfiles";
   private static String loggingName = "makejmlrbook.log";

   private static final int MAX_RECENT_FILES=10;

   private MakeJmlrBookGUI app;

   private File recentFile;

   private File loggingFile;

   private static File resourceDir;

   private Vector<String> recentList;
}
