/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui.newbook;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.MalformedURLException;
import java.nio.file.Path;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.regex.*;
import java.util.Calendar;
import java.text.Normalizer;
import java.nio.charset.Charset;
import java.nio.charset.IllegalCharsetNameException;

import javax.swing.*;
import javax.swing.table.TableModel;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableColumn;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.dickimawbooks.texparserlib.*;
import com.dickimawbooks.texparserlib.latex.*;
import com.dickimawbooks.texparserlib.latex.jmlr.*;
import com.dickimawbooks.texparserlib.latex.inputenc.InputEncSty;
import com.dickimawbooks.texparserlib.latex2latex.*;
import com.dickimawbooks.texparserlib.bib.*;
import com.dickimawbooks.texparserlib.html.L2HStringConverter;

import com.dickimawbooks.makejmlrbookgui.*;

public class ProcessPanel extends JPanel
  implements NewBookPanel,Runnable,ActionListener,LaTeXApp
{
   public ProcessPanel(NewBookDialog dialog)
   {
      super(new BorderLayout());
      this.dialog = dialog;

      errMessageBuffer = new StringBuffer();

      messageArea = new MessagePane();

      add(new JScrollPane(messageArea), BorderLayout.CENTER);

      buttonPanel = new JPanel();

      buttonPanel.add(JMLRResources.createActionButton("newbook.process",
        "copy", this, 
        KeyStroke.getKeyStroke(KeyEvent.VK_C, InputEvent.CTRL_MASK)));

      buttonPanel.add(JMLRResources.createActionButton("newbook.process",
        "view", this, 
        KeyStroke.getKeyStroke(KeyEvent.VK_E, InputEvent.CTRL_MASK)));

      buttonPanel.add(JMLRResources.createActionButton("newbook.process",
        "finish", this, null));

      add(buttonPanel, BorderLayout.SOUTH);
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("copy"))
      {
         messageArea.selectAll();
         messageArea.copy();
      }
      else if (action.equals("finish"))
      {
         dialog.finish(bookFile);
      }
      else if (action.equals("view"))
      {
         if (bookFile == null)
         {
            JMLRResources.error(dialog,
              MakeJmlrBookGUI.getLabel("error.load_failed"));
         }
         else if (bookFile.exists())
         {
            try
            {
               dialog.viewFile(bookFile);
            }
            catch (MalformedURLException e)
            {
               JMLRResources.error(dialog, e);
            }
         }
         else
         {
            JMLRResources.error(dialog,
              MakeJmlrBookGUI.getLabelWithValue("error.file_not_found",
                 bookFile.getAbsolutePath()));
         }
      }
   }

   public boolean canMoveForwards() {return false;}

   public boolean hasNextButton() {return false;}

   public boolean hasBackButton() {return false;}

   public void updatePanel()
   {
      buttonPanel.setVisible(false);

      errMessageBuffer.setLength(0);
      messageArea.clearMessages();

      Thread thread = new Thread(this);
      thread.start();
      thread = null;
   }

   public void copyFile(File src, File dest)
     throws IOException
   {
      dialog.verboseCopyFile(src, dest);
   }

   public void epstopdf(File epsFile, File pdfFile)
     throws IOException
   {
      dialog.epstopdf(epsFile, pdfFile);
   }

   public void wmftoeps(File wmfFile, File epsFile)
     throws IOException
   {
      dialog.wmftoeps(wmfFile, epsFile);
   }

   public String kpsewhich(String name)
     throws IOException
   {
      return dialog.kpsewhich(name);
   }

   public void progress(int progress)
   {
   }

   public boolean isReadAccessAllowed(TeXPath path)
   {
      return isReadAccessAllowed(path.getFile());
   }

   public boolean isReadAccessAllowed(File file)
   {
      return file.canRead();
   }

   public boolean isWriteAccessAllowed(TeXPath path)
   {
      return isWriteAccessAllowed(path.getFile());
   }

   public boolean isWriteAccessAllowed(File file)
   {
      return file.canWrite();
   }

   public String requestUserInput(String message)
     throws IOException
   {
      return javax.swing.JOptionPane.showInputDialog(dialog, message);
   }

   private String taggify(String tag) throws IOException
   {
      if (tag.indexOf("\\") != -1)
      {
         L2HStringConverter listener = new L2HStringConverter(
           new BookTeXApp(dialog.getApplication()));

         tag = listener.convert(tag, false);
      }

      return Normalizer.normalize(tag,
               Normalizer.Form.NFD).replaceAll(
             "\\p{M}", "").toLowerCase().replaceAll("[^a-z\\-]+", "");
   }

   private void createTags() throws IOException
   {
      BookTeXApp texApp = new BookTeXApp(dialog.getApplication());

      attrMap = new Hashtable<File,ImportFileAttributes>();

      String year = String.format("%02d", (dialog.getYear()%100));

      File file = dialog.getPrefaceFile();

      if (hasPreface)
      {
         Contributor editor = editors.firstElement();

         String tag = L2HStringConverter.convert(texApp, editor.getSurname(),
          false);

         tag = taggify(tag);

         if (file == null)
         {
            file = new File(frontMatterDir, PREFACE+".tex");
         }

         attrMap.put(file, new ImportFileAttributes(this, frontMatterDir, 
           tag+year));
      }

      if (partConfig == BookPartPanel.PARTS_NO)
      {
         createTags(dialog.getMainImportPapers().getPapers(), texApp, year);
      }
      else
      {
         Vector<PartListPanel> partPanels = dialog.getPartLists();

         for (PartListPanel partPanel : partPanels)
         {
            createTags(partPanel.getImportPapersPanel().getPapers(), texApp, year);
         }
      }
   }

   private void createTags(Vector<ImportPaperPanel> papers, TeXApp texApp, String year)
    throws IOException
   {
      for (ImportPaperPanel paper: papers)
      {
         File file = paper.getFile();

         PreambleParser preambleParser = new PreambleParser(texApp);
         TeXParser texParser = preambleParser.parse(file);

         Charset charset = null;

         String encoding = preambleParser.getInputEncoding();

         if (encoding != null && !encoding.equals("utf8"))
         {
            charset = InputEncSty.getCharSet(encoding);
         }

         LaTeXFile docCls = preambleParser.getDocumentClass();

         String tag = null;

         Vector<Group> authors = null;

         if (docCls instanceof JmlrCls)
         {
            authors = ((JmlrCls)docCls).getAuthors();

            if (authors == null)
            {
               TeXObjectList author = preambleParser.getAuthor();

               if (author != null)
               {
                  author.process(texParser);
               }

               authors = ((JmlrCls)docCls).getAuthors();
            }

         }
         else
         {
            LaTeXSty sty = preambleParser.getLoadedPackage("jmlr2e");

            if (sty instanceof Jmlr2eSty)
            {
               authors = ((Jmlr2eSty)sty).getAuthors();

               if (authors == null)
               {
                  TeXObjectList author = preambleParser.getAuthor();

                  if (author != null)
                  {
                     author.process(texParser);
                  }

                  authors = ((Jmlr2eSty)sty).getAuthors();
               }
            }
         }

         Group author = (authors==null ? null : authors.firstElement());

         if (author != null)
         {
            TeXObjectList list = new TeXObjectList();

            for (int j = author.size()-1; j >= 0; j--)
            {
               TeXObject object = author.get(j);

               if (object instanceof WhiteSpace)
               {
                  break;
               }

               list.push(object);
            }

            tag = list.expandfully(texParser).format();
         }

         if (tag == null)
         {
            tag = file.getName();
         }

         tag = taggify(tag)+year;

         int suffix = 0;

         for (Enumeration<File> e = attrMap.keys(); e.hasMoreElements();)
         {
            File key = e.nextElement();
            ImportFileAttributes value = attrMap.get(key);
            String valTag = value.getTag();

            if (tag.equals(value.getTag()))
            {
               value.setTag(tag+"a");
               suffix = (int)'a';

               break;
            }

            if (valTag.startsWith(tag))
            {
               String end = valTag.substring(tag.length()+1);

               if (end.length() == 1)
               {
                  char c = end.charAt(0);

                  if (c > suffix)
                  {
                     suffix = c;
                  }
               }
            }
         }

         if (suffix != 0)
         {
            tag = String.format("%s%c", tag, (char)(suffix+1));
         }

         ImportFileAttributes attr = new ImportFileAttributes(this, tag,
           charset, preambleParser.getLoadedPackages(), authors);

         paper.setAttributes(attr);
         attrMap.put(file, attr);
      }
   }

   public void run()
   {
      reader = null;
      writer = null;
      currentFile = null;
      bookDir = dialog.getBookDirectory();
      base = bookDir.getName();
      prefaceFile = null;
      forewordFile = null;

      title = dialog.getBookTitle();
      shortTitle = dialog.getBookShortTitle();
      editors = dialog.getEditors();

      styNames = new Hashtable<String,PackageInfo>();

      message("h1",
        MakeJmlrBookGUI.getLabelWithValue("newbook.process.title", base));

      hasPreface = dialog.hasPreface();
      hasForeword = dialog.hasForeword();
      hasAppendices = dialog.hasAppendices();

      frontMatterDir = new File(bookDir, FRONTMATTER);

      try
      {
         createTags();
      }
      catch (IOException e)
      {
         error(e.getMessage());
         MakeJmlrBookGUI.debug(e);
      }

      if (hasPreface || hasForeword)
      {
         doFrontMatter();
      }

      partConfig = dialog.getPartConfig();

      doMainMatter();

      if (hasAppendices)
      {
         doAppendices();
      }

      writeBookFile();

      if (errMessageBuffer.length() > 0)
      {
         JMLRResources.error(dialog, 
            MakeJmlrBookGUI.getLabelWithValue(
              "newbook.completed_with_errors",
              errMessageBuffer.toString()));
      }
      else
      {
         JOptionPane.showMessageDialog(dialog, 
           MakeJmlrBookGUI.getLabel("newbook.completed"),
           MakeJmlrBookGUI.getLabel("newbook.completed.title"),
           JOptionPane.INFORMATION_MESSAGE);
      }

      buttonPanel.setVisible(true);

      message(MakeJmlrBookGUI.getLabel("message.done"));
      dialog.setAppInfo(MakeJmlrBookGUI.getLabel("message.done"));
   }

   private void doFrontMatter()
   {
      message("h2", MakeJmlrBookGUI.getLabel("newbook.frontmatter.label"));

      message(MakeJmlrBookGUI.getLabelWithValue("newbook.process.mkdir",
        frontMatterDir.getAbsolutePath()));

      if (!frontMatterDir.mkdir())
      {
         error(MakeJmlrBookGUI.getLabelWithValue(
           "newbook.process.cant_mkdir",
        frontMatterDir.getAbsolutePath()));

         hasPreface = false;
         hasForeword = false;
      }

      if (hasPreface)
      {
         doPreface();
      }

      if (hasForeword)
      {
         doForeword();
      }
   }

   private void doPreface()
   {
      message("h3",
        MakeJmlrBookGUI.getLabel("newbook.frontmatter.preface.label"));

      prefaceFile = dialog.getPrefaceFile();
      currentFile = prefaceFile;

      try
      {
         String prefaceTag = null;

         if (prefaceFile == null)
         {
            prefaceFile = new File(frontMatterDir, PREFACE+".tex");

            message(MakeJmlrBookGUI.getLabelWithValue("message.writing",
               prefaceFile.getAbsolutePath()));

            writer = new PrintWriter(prefaceFile);

            Contributor first = editors.firstElement();

            prefaceTag = attrMap.get(prefaceFile).getTag();

            writer.println("\\begin{preface}["+prefaceTag+"]");
            writer.println();
            writer.println("Insert preface text here.");
            writer.println();

            writer.println("\\begin{signoff}{Insert Date Here}");

            for (Contributor editor : editors)
            {
               String surname = editor.getSurname();

               if (surname.contains(" "))
               {
                  surname = "{"+surname+"}";
               }

               writer.println("% Editor details: ");
               writer.println("\\Editor{" + contribFormat.format(editor) +"\\\\");
               writer.println("Insert Address Here\\\\");
               writer.println("\\mailto{insert@email.here}}");
            }

            writer.println("\\end{signoff}");
            writer.println("\\end{preface}");

            writer.close();
            writer = null;
         }
         else
         {
            prefaceTag = attrMap.get(prefaceFile).getTag();

            PrefaceParser prefParser = new PrefaceParser(this,
               frontMatterDir, editors, prefaceTag);

            TeXParser parser = new TeXParser(prefParser);

            parser.parse(prefaceFile);

            prefaceFile = new File(frontMatterDir, prefaceFile.getName());
         }

         if (!dialog.isCiML())
         {
            File file = new File(frontMatterDir, prefaceTag+".tex");

            writer = new PrintWriter(file);

            writer.println("% "
              + MakeJmlrBookGUI.getLabel("newbook.comment.wcporpmlr"));

            writer.println("\\documentclass[pmlr]{jmlrbook}");

            writer.println();
            writer.print("\\title");

            if (!shortTitle.isEmpty())
            {
               writer.print("["+shortTitle+"]");
            }

            writer.println(String.format("{%s}", title));

            int n = editors.size();

            writer.print("\\author{");

            Contributor firstEd = editors.firstElement();
            writer.print(contribFormat.format(firstEd));

            if (n > 1)
            {
               writer.print(" et al.");
            }

            writer.println("}");
            writer.println();

            writer.println("\\begin{document}");
            writer.println(String.format("\\input{%s}", prefaceFile.getName()));
            writer.println("\\end{document}");

            writer.close();
            writer = null;
         }
      }
      catch (IOException e)
      {
         prefaceFile = null;
         hasPreface = false;
         JMLRResources.error(dialog, e);
      }
      finally
      {
         if (writer != null)
         {
            writer.close();
            writer = null;
         }
      }
   }

   private void doForeword()
   {
      message("h3",
        MakeJmlrBookGUI.getLabel("newbook.frontmatter.foreword.label"));

      forewordFile = dialog.getForewordFile();
      currentFile = forewordFile;

      try
      {
         if (forewordFile == null)
         {
            forewordFile = new File(frontMatterDir, FOREWORD+".tex");

            message(MakeJmlrBookGUI.getLabelWithValue("message.writing",
               forewordFile.getAbsolutePath()));

            writer = new PrintWriter(forewordFile);

            writer.println("\\chapter{Foreword}");

            writer.println();
            writer.println("Insert foreword text here.");
            writer.println();
            writer.println("\\begin{authorsignoff}");
            writer.println("\\Author{Insert Name Here\\\\");
            writer.println("Insert Affiliation Here}");
            writer.println("\\end{authorsignoff}");
         }
         else
         {
            ForewordParser forewordParser = new ForewordParser(this,
               frontMatterDir);

            TeXParser parser = new TeXParser(forewordParser);

            parser.parse(forewordFile);

            forewordFile = new File(frontMatterDir, forewordFile.getName());
         }
      }
      catch (IOException e)
      {
         forewordFile = null;
         hasForeword = false;
         JMLRResources.error(dialog, e);
      }
      finally
      {
         if (writer != null)
         {
            writer.close();
            writer = null;
         }
      }

   }

   private void doMainMatter()
   {
      message("h2", MakeJmlrBookGUI.getLabel("newbook.articles.main"));

      if (partConfig == BookPartPanel.PARTS_NO)
      {
         ImportPapersPanel mainImportPanel = dialog.getMainImportPapers();

         doImportPapers(mainImportPanel.getPapers());
      }
      else
      {
         Vector<PartListPanel> partPanels = dialog.getPartLists();

         for (PartListPanel partPanel : partPanels)
         {
            message("h2", partPanel.getPartTitle());

            ImportPapersPanel impPanel = partPanel.getImportPapersPanel();

            doImportPapers(impPanel.getPapers());
         }
      }
   }

   private void doAppendices()
   {
      message("h2", MakeJmlrBookGUI.getLabel("newbook.articles.appendix"));

      ImportPapersPanel impPanel = dialog.getAppendixImportPapers();

      doImportPapers(impPanel.getPapers());
   }

   public PackageInfo addPackage(String styName, String options)
   {
      PackageInfo pkgInfo;

      if (styName.equals("algpseudocode"))
      {
         styNames.remove("algorithmicx");
      }
      else if (styName.equals("algorithmicx"))
      {
         if (styNames.containsKey("algpseudocode"))
         {
            return null;
         }
      }

      if (styNames.containsKey(styName))
      {
         pkgInfo = styNames.get(styName);

         pkgInfo.addOption(options);
      }
      else
      {
         pkgInfo = new PackageInfo(styName, options, getPackageLocation(styName));

         styNames.put(styName, pkgInfo);
      }

      return pkgInfo;
   }

   private void doImportPapers(Vector<ImportPaperPanel> papers)
   {
      java.io.FileFilter suppFilter = new java.io.FileFilter()
      {
         public boolean accept(File file)
         {
            if (file.isDirectory()) return false;

            return file.getName().contains("-supp.");
         }
      };

      Pattern p = Pattern.compile("(.+)-supp\\.(.+)");

      for (ImportPaperPanel paper: papers)
      {
         File file = paper.getFile();
         currentFile = file;

         ImportFileAttributes attr = paper.getAttributes();

         if (attr == null)
         {
            attr = attrMap.get(file);
            paper.setAttributes(attr);
         }

         File impFile = paper.getToImportFile();

         if (impFile == null)
         {
            continue;
         }

         File dir = impFile.getParentFile();

         if (!dir.exists())
         {
            dir.mkdir();
         }

         try
         {
            // copy over any supplemental files

            File[] fileList = file.getParentFile().listFiles(suppFilter);

            String tag = attr.getTag();

            for (File suppFile : fileList)
            {
               String name = suppFile.getName();

               Matcher m = p.matcher(name);

               if (m.matches())
               {
                  String prefix = m.group(1);
                  String ext = m.group(2);

                  String destName = tag+"-supp."+ext;

                  File destFile = new File(dir, destName);
                  int suffix = 1;

                  while (destFile.exists())
                  {
                     destName = String.format("%s%d-supp.%s", tag, suffix++, ext);
                     destFile = new File(dir, destName);
                  }

                  copyFile(suppFile, destFile);
               }
            }

            ArticleParser artParser = new ArticleParser(this,
               paper.getPackages(), impFile, !dialog.isCiML());

            TeXParser parser = new TeXParser(artParser);

            artParser.usepackage(null, "graphicx", false);

            parser.parse(file, paper.getCharSet());
         }
         catch (IOException e)
         {
            error(e);
         }
      }
   }

   private void writeFontFile()
     throws IOException
   {
      File fontFile = new File(bookDir, "bookfonts.tex");

      message(MakeJmlrBookGUI.getLabelWithValue("message.writing",
          fontFile.getAbsolutePath()));

      PrintWriter out = null;

      try
      {
         out = new PrintWriter(fontFile);

         out.println("% Book fonts");

         out.println(dialog.getBookFonts());
      }
      finally
      {
         if (out != null)
         {
            out.close();
            out = null;
         }
      }
   }

   private void writeBookFile()
   {
      try
      {
         writeFontFile();

         bookFile = new File(bookDir, base+".tex");
         Path bookPath = bookDir.toPath();

         message(MakeJmlrBookGUI.getLabelWithValue("message.writing",
            bookFile.getAbsolutePath()));
         writer = new PrintWriter(bookFile);

         writer.println("% "+MakeJmlrBookGUI.getLabelWithValue(
           "newbook.comment.prehyperref", "\\jmlrprehyperref"));
         writer.println();
         writer.println("\\def\\jmlrprehyperref{%");
         writer.println("% "
           + MakeJmlrBookGUI.getLabel("newbook.comment.import_sty"));

         boolean algorithmic = false;
         boolean algorithmicx = false;

         for (Enumeration<String> en=styNames.keys(); en.hasMoreElements();)
         {
            String key = en.nextElement();

            PackageInfo pkgInfo = styNames.get(key);

            if (key.equals("inputenc"))
            {
               if (!pkgInfo.hasOption("utf8"))
               {
                  message(MakeJmlrBookGUI.getLabelWithValue(
                    "error.inputenc.notutf8", pkgInfo.getOptions()));
               }

               continue;
            }

            if (key.equals("fontenc"))
            {
               if (!pkgInfo.hasOption("T1"))
               {
                  message(MakeJmlrBookGUI.getLabelWithValue(
                    "error.inputenc.notT1", pkgInfo.getOptions()));
               }

               continue;
            }

            if (key.equals("algpseudocode") || key.equals("algorithmicx"))
            {
               algorithmicx = true;
               writer.println(" % "
                + MakeJmlrBookGUI.getLabelWithValue("newbook.comment.patch_sty",
                     key));
               writer.println(" % "
                +MakeJmlrBookGUI.getLabelWithValue(
                    "newbook.comment.save_sty_cs", "algorithm2e"));
               writer.println(" \\let\\ALGOFor\\For");
               writer.println(" \\let\\ALGOForEach\\ForEach");
               writer.println(" \\let\\ALGOForAll\\ForAll");
               writer.println(" \\let\\ALGOWhile\\While");
               writer.println(" \\let\\ALGOReturn\\Return");
               writer.println(" \\let\\ALGOBegin\\Begin");
               writer.println(" \\let\\ALGOIf\\If");
               writer.println(" \\let\\ALGOElse\\Else");
               writer.println(" \\let\\ALGOSwitch\\Switch");
               writer.println(" \\let\\ALGOCase\\Case");
               writer.println(" \\let\\ALGOOther\\Other");

               writer.println(" \\def\\switchtoALGO{%");
               writer.println("  \\let\\For\\ALGOFor");
               writer.println("  \\let\\ForEach\\ALGOForEach");
               writer.println("  \\let\\ForAll\\ALGOForAll");
               writer.println("  \\let\\While\\ALGOWhile");
               writer.println("  \\let\\Return\\ALGOReturn");
               writer.println("  \\let\\Begin\\ALGOBegin");
               writer.println("  \\let\\If\\ALGOIf");
               writer.println("  \\let\\Else\\ALGOElse");
               writer.println("  \\let\\Switch\\ALGOSwitch");
               writer.println("  \\let\\Case\\ALGOCase");
               writer.println("  \\let\\Other\\ALGOOther");
               writer.println(" }");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValues(
                 "newbook.comment.undef_sty_cs", "algorithm2e", key));

               writer.println(" \\undef\\For");
               writer.println(" \\undef\\ForEach");
               writer.println(" \\undef\\ForAll");
               writer.println(" \\undef\\While");
               writer.println(" \\undef\\Return");
               writer.println(" \\undef\\Begin");
               writer.println(" \\undef\\If");
               writer.println(" \\undef\\Else");
               writer.println(" \\undef\\Switch");
               writer.println(" \\undef\\Case");
               writer.println(" \\undef\\Other");

               if (algorithmic)
               {
                  writer.println(" % "+MakeJmlrBookGUI.getLabelWithValues(
                     "newbook.comment.undef_sty_cs", "algorithmic", key));

                  writer.println(" \\undef\\algorithmicindent");
                  writer.println(" \\undef\\algorithmic");
                  writer.println(" \\undef\\endalgorithmic");
                  writer.println(" \\undef\\algorithmiccomment");
                  writer.println(" \\undef\\algorithmicrequire");
                  writer.println(" \\undef\\algorithmicensure");
                  writer.println(" \\undef\\algorithmicend");
                  writer.println(" \\undef\\algorithmicelse");
                  writer.println(" \\undef\\algorithmicfor");
                  writer.println(" \\undef\\algorithmicforall");
                  writer.println(" \\undef\\algorithmicdo");
                  writer.println(" \\undef\\algorithmicwhile");
                  writer.println(" \\undef\\algorithmicloop");
                  writer.println(" \\undef\\algorithmicrepeat");
                  writer.println(" \\undef\\algorithmicuntil");
                  writer.println(" \\undef\\algorithmicreturn");
                  writer.println(" \\undef\\algorithmicif");
                  writer.println(" \\undef\\algorithmicthen");
               }

               writer.print(" ");
               pkgInfo.write(writer);

               writer.println(" \\let\\ALGPSEUDOCODEalgorithmic\\algorithmic");
               writer.println(" \\let\\endALGPSEUDOCODEalgorithmic\\endalgorithmic");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicindent\\algorithmicindent");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmiccomment\\algorithmiccomment");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicrequire\\algorithmicrequire");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicensure\\algorithmicensure");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicend\\algorithmicend");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicelse\\algorithmicelse");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicfor\\algorithmicfor");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicforall\\algorithmicforall");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicdo\\algorithmicdo");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicwhile\\algorithmicwhile");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicloop\\algorithmicloop");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicrepeat\\algorithmicrepeat");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicuntil\\algorithmicuntil");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicreturn\\algorithmicreturn");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicif\\algorithmicif");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicthen\\algorithmicthen");
               writer.println(" \\let\\ALGPSEUDOCODEFor\\For");
               writer.println(" \\let\\ALGPSEUDOCODEForEach\\ForEach");
               writer.println(" \\let\\ALGPSEUDOCODEForAll\\ForAll");
               writer.println(" \\let\\ALGPSEUDOCODEWhile\\While");
               writer.println(" \\let\\ALGPSEUDOCODEReturn\\Return");
               writer.println(" \\let\\ALGPSEUDOCODEBegin\\Begin");
               writer.println(" \\let\\ALGPSEUDOCODEIf\\If");
               writer.println(" \\let\\ALGPSEUDOCODEElse\\Else");
               writer.println(" \\let\\ALGPSEUDOCODESwitch\\Switch");
               writer.println(" \\let\\ALGPSEUDOCODECase\\Case");
               writer.println(" \\let\\ALGPSEUDOCODEOther\\Other");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));
               writer.println(" \\def\\switchtoALGPSEUDOCODE{%");
               writer.println("  \\let\\algorithmic\\ALGPSEUDOCODEalgorithmic");
               writer.println("  \\let\\endalgorithmic\\endALGPSEUDOCODEalgorithmic");
               writer.println("  \\let\\algorithmicindent\\ALGPSEUDOCODEalgorithmicindent");
               writer.println("  \\let\\algorithmiccomment\\ALGPSEUDOCODEalgorithmiccomment");
               writer.println("  \\let\\algorithmicrequire\\ALGPSEUDOCODEalgorithmicrequire");
               writer.println("  \\let\\algorithmicensure\\ALGPSEUDOCODEalgorithmicensure");
               writer.println("  \\let\\algorithmicend\\ALGPSEUDOCODEalgorithmicend");
               writer.println("  \\let\\algorithmicelse\\ALGPSEUDOCODEalgorithmicelse");
               writer.println("  \\let\\algorithmicfor\\ALGPSEUDOCODEalgorithmicfor");
               writer.println("  \\let\\algorithmicforall\\ALGPSEUDOCODEalgorithmicforall");
               writer.println("  \\let\\algorithmicdo\\ALGPSEUDOCODEalgorithmicdo");
               writer.println("  \\let\\algorithmicwhile\\ALGPSEUDOCODEalgorithmicwhile");
               writer.println("  \\let\\algorithmicloop\\ALGPSEUDOCODEalgorithmicloop");
               writer.println("  \\let\\algorithmicrepeat\\ALGPSEUDOCODEalgorithmicrepeat");
               writer.println("  \\let\\algorithmicuntil\\ALGPSEUDOCODEalgorithmicuntil");
               writer.println("  \\let\\algorithmicreturn\\ALGPSEUDOCODEalgorithmicreturn");
               writer.println("  \\let\\algorithmicif\\ALGPSEUDOCODEalgorithmicif");
               writer.println("  \\let\\algorithmicthen\\ALGPSEUDOCODEalgorithmicthen");
               writer.println(" \\let\\For\\ALGPSEUDOCODEFor");
               writer.println(" \\let\\ForEach\\ALGPSEUDOCODEForEach");
               writer.println(" \\let\\ForAll\\ALGPSEUDOCODEForAll");
               writer.println(" \\let\\While\\ALGPSEUDOCODEWhile");
               writer.println(" \\let\\Return\\ALGPSEUDOCODEReturn");
               writer.println(" \\let\\Begin\\ALGPSEUDOCODEBegin");
               writer.println(" \\let\\If\\ALGPSEUDOCODEIf");
               writer.println(" \\let\\Else\\ALGPSEUDOCODEElse");
               writer.println(" \\let\\Switch\\ALGPSEUDOCODESwitch");
               writer.println(" \\let\\Case\\ALGPSEUDOCODECase");
               writer.println(" \\let\\Other\\ALGPSEUDOCODEOther");
               writer.println(" }%");

            }
            else if (key.equals("algorithmic"))
            {
               algorithmic = true;
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.patch_sty", key));

               if (algorithmic)
               {
                  writer.println(" % "+MakeJmlrBookGUI.getLabelWithValues(
                    "ewbook.comment.undef_sty_cs", "algorithmicx", key));
                  writer.println(" \\undef\\algorithmicindent");
                  writer.println(" \\undef\\algorithmiccomment");
                  writer.println(" \\undef\\algorithmicrequire");
                  writer.println(" \\undef\\algorithmicensure");
                  writer.println(" \\undef\\algorithmicend");
                  writer.println(" \\undef\\algorithmicelse");
                  writer.println(" \\undef\\algorithmicfor");
                  writer.println(" \\undef\\algorithmicforall");
                  writer.println(" \\undef\\algorithmicdo");
                  writer.println(" \\undef\\algorithmicwhile");
                  writer.println(" \\undef\\algorithmicloop");
                  writer.println(" \\undef\\algorithmicrepeat");
                  writer.println(" \\undef\\algorithmicuntil");
                  writer.println(" \\undef\\algorithmicreturn");
                  writer.println(" \\undef\\algorithmicif");
                  writer.println(" \\undef\\algorithmicthen");
                  writer.println(" \\undef\\algorithmicendif");
                  writer.println(" \\undef\\algorithmicendfor");
                  writer.println(" \\undef\\algorithmicendwhile");
                  writer.println(" \\undef\\algorithmicendloop");
                  writer.println(" \\undef\\algorithmicelsif");
                  writer.println(" \\undef\\algorithmicprint");
                  writer.println(" \\undef\\algorithmicand");
                  writer.println(" \\undef\\algorithmicor");
                  writer.println(" \\undef\\algorithmicxor");
                  writer.println(" \\undef\\algorithmicnot");
                  writer.println(" \\undef\\algorithmicto");
                  writer.println(" \\undef\\algorithmicinputs");
                  writer.println(" \\undef\\algorithmicoutputs");
                  writer.println(" \\undef\\algorithmicglobals");
                  writer.println(" \\undef\\algorithmicbody");
                  writer.println(" \\undef\\algorithmictrue");
                  writer.println(" \\undef\\algorithmicfalse");
               }

               writer.print(" ");

               pkgInfo.write(writer);

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.save_sty_cs", key));
               writer.println("\\letcs{\\ALGORITHMICialgorithmic}{\\string\\algorithmic}%");
               writer.println(" \\let\\ALGORITHMICalgorithmic\\algorithmic");
               writer.println(" \\let\\endALGORITHMICalgorithmic\\endalgorithmic");
               writer.println(" \\let\\ALGORITHMICalgorithmicindent\\algorithmicindent");
               writer.println(" \\let\\ALGORITHMICalgorithmiccomment\\algorithmiccomment");
               writer.println(" \\let\\ALGORITHMICalgorithmicrequire\\algorithmicrequire");
               writer.println(" \\let\\ALGORITHMICalgorithmicensure\\algorithmicensure");
               writer.println(" \\let\\ALGORITHMICalgorithmicend\\algorithmicend");
               writer.println(" \\let\\ALGORITHMICalgorithmicelse\\algorithmicelse");
               writer.println(" \\let\\ALGORITHMICalgorithmicfor\\algorithmicfor");
               writer.println(" \\let\\ALGORITHMICalgorithmicforall\\algorithmicforall");
               writer.println(" \\let\\ALGORITHMICalgorithmicdo\\algorithmicdo");
               writer.println(" \\let\\ALGORITHMICalgorithmicwhile\\algorithmicwhile");
               writer.println(" \\let\\ALGORITHMICalgorithmicloop\\algorithmicloop");
               writer.println(" \\let\\ALGORITHMICalgorithmicrepeat\\algorithmicrepeat");
               writer.println(" \\let\\ALGORITHMICalgorithmicuntil\\algorithmicuntil");
               writer.println(" \\let\\ALGORITHMICalgorithmicreturn\\algorithmicreturn");
               writer.println(" \\let\\ALGORITHMICalgorithmicif\\algorithmicif");
               writer.println(" \\let\\ALGORITHMICalgorithmicthen\\algorithmicthen");
               writer.println(" \\let\\ALGORITHMICalgorithmicendif\\algorithmicendif");
               writer.println(" \\let\\ALGORITHMICalgorithmicendfor\\algorithmicendfor");
               writer.println(" \\let\\ALGORITHMICalgorithmicendwhile\\algorithmicendwhile");
               writer.println(" \\let\\ALGORITHMICalgorithmicendloop\\algorithmicendloop");
               writer.println(" \\let\\ALGORITHMICalgorithmicelsif\\algorithmicelsif");
               writer.println(" \\let\\ALGORITHMICalgorithmicprint\\algorithmicprint");
               writer.println(" \\let\\ALGORITHMICalgorithmicand\\algorithmicand");
               writer.println(" \\let\\ALGORITHMICalgorithmicor\\algorithmicor");
               writer.println(" \\let\\ALGORITHMICalgorithmicxor\\algorithmicxor");
               writer.println(" \\let\\ALGORITHMICalgorithmicnot\\algorithmicnot");
               writer.println(" \\let\\ALGORITHMICalgorithmicto\\algorithmicto");
               writer.println(" \\let\\ALGORITHMICalgorithmicinputs\\algorithmicinputs");
               writer.println(" \\let\\ALGORITHMICalgorithmicoutputs\\algorithmicoutputs");
               writer.println(" \\let\\ALGORITHMICalgorithmicglobals\\algorithmicglobals");
               writer.println(" \\let\\ALGORITHMICalgorithmicbody\\algorithmicbody");
               writer.println(" \\let\\ALGORITHMICalgorithmictrue\\algorithmictrue");
               writer.println(" \\let\\ALGORITHMICalgorithmicfalse\\algorithmicfalse");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));
               writer.println(" \\def\\switchtoALGORITHMIC{");
               writer.println("  \\cslet{\\string\\algorithmic}{\\ALGORITHMICialgorithmic}%");
               writer.println("  \\let\\algorithmic\\ALGORITHMICalgorithmic");
               writer.println("  \\let\\endalgorithmic\\endALGORITHMICalgorithmic");
               writer.println("  \\let\\algorithmicindent\\ALGORITHMICalgorithmicindent");
               writer.println("  \\let\\algorithmiccomment\\ALGORITHMICalgorithmiccomment");
               writer.println("  \\let\\algorithmicrequire\\ALGORITHMICalgorithmicrequire");
               writer.println("  \\let\\algorithmicensure\\ALGORITHMICalgorithmicensure");
               writer.println("  \\let\\algorithmicend\\ALGORITHMICalgorithmicend");
               writer.println("  \\let\\algorithmicelse\\ALGORITHMICalgorithmicelse");
               writer.println("  \\let\\algorithmicfor\\ALGORITHMICalgorithmicfor");
               writer.println("  \\let\\algorithmicforall\\ALGORITHMICalgorithmicforall");
               writer.println("  \\let\\algorithmicdo\\ALGORITHMICalgorithmicdo");
               writer.println("  \\let\\algorithmicwhile\\ALGORITHMICalgorithmicwhile");
               writer.println("  \\let\\algorithmicloop\\ALGORITHMICalgorithmicloop");
               writer.println("  \\let\\algorithmicrepeat\\ALGORITHMICalgorithmicrepeat");
               writer.println("  \\let\\algorithmicuntil\\ALGORITHMICalgorithmicuntil");
               writer.println("  \\let\\algorithmicreturn\\ALGORITHMICalgorithmicreturn");
               writer.println("  \\let\\algorithmicif\\ALGORITHMICalgorithmicif");
               writer.println("  \\let\\algorithmicthen\\ALGORITHMICalgorithmicthen");
               writer.println("  \\let\\algorithmicendif\\ALGORITHMICalgorithmicendif");
               writer.println("  \\let\\algorithmicendfor\\ALGORITHMICalgorithmicendfor");
               writer.println("  \\let\\algorithmicendwhile\\ALGORITHMICalgorithmicendwhile");
               writer.println("  \\let\\algorithmicendloop\\ALGORITHMICalgorithmicendloop");
               writer.println("  \\let\\algorithmicelsif\\ALGORITHMICalgorithmicelsif");
               writer.println("  \\let\\algorithmicprint\\ALGORITHMICalgorithmicprint");
               writer.println("  \\let\\algorithmicand\\ALGORITHMICalgorithmicand");
               writer.println("  \\let\\algorithmicor\\ALGORITHMICalgorithmicor");
               writer.println("  \\let\\algorithmicxor\\ALGORITHMICalgorithmicxor");
               writer.println("  \\let\\algorithmicnot\\ALGORITHMICalgorithmicnot");
               writer.println("  \\let\\algorithmicto\\ALGORITHMICalgorithmicto");
               writer.println("  \\let\\algorithmicinputs\\ALGORITHMICalgorithmicinputs");
               writer.println("  \\let\\algorithmicoutputs\\ALGORITHMICalgorithmicoutputs");
               writer.println("  \\let\\algorithmicglobal\\ALGORITHMICalgorithmicglobal");
               writer.println("  \\let\\algorithmicbody\\ALGORITHMICalgorithmicbody");
               writer.println("  \\let\\algorithmictrue\\ALGORITHMICalgorithmictrue");
               writer.println("  \\let\\algorithmicfalse\\ALGORITHMICalgorithmicfalse");
               writer.println("  \\undef\\REQUIRE");
               writer.println("  \\undef\\ENSURE");
               writer.println("  \\undef\\PRINT");
               writer.println("  \\undef\\TRUE");
               writer.println("  \\undef\\FALSE");
               writer.println("  \\undef\\AND");
               writer.println("  \\undef\\OR");
               writer.println("  \\undef\\NOT");
               writer.println("  \\undef\\STATE");
               writer.println("  \\undef\\STMT");
               writer.println("  \\undef\\COMMENT");
               writer.println("  \\undef\\INPUTS");
               writer.println("  \\undef\\OUTPUTS");
               writer.println("  \\undef\\ENDOUTPUTS");
               writer.println("  \\undef\\GLOBALS");
               writer.println("  \\undef\\BODY");
               writer.println("  \\undef\\ENDBODY");
               writer.println("  \\undef\\IF");
               writer.println("  \\undef\\ENDIF");
               writer.println("  \\undef\\ELSE");
               writer.println("  \\undef\\ELSIF");
               writer.println("  \\undef\\FOR");
               writer.println("  \\undef\\FORALL");
               writer.println("  \\undef\\ENDFOR");
               writer.println("  \\undef\\WHILE");
               writer.println("  \\undef\\ENDWHILE");
               writer.println("  \\undef\\REPEAT");
               writer.println("  \\undef\\UNTIL");
               writer.println("  \\undef\\LOOP");
               writer.println("  \\undef\\ENDLOOP");
               writer.println("  \\undef\\RETURN");

               writer.println(" }");

            }
            else if (key.equals("mdwlist"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.patch_sty", key));
               pkgInfo.write(writer);
               writer.println("  \\csletcs{mdwenumerate*}{enumerate*}");
               writer.println("  \\csletcs{endmdwenumerate*}{endenumerate*}");
               writer.println("  \\csundef{enumerate*}");
               writer.println("  \\csundef{endenumerate*}");
            }
            else if (key.equals("amssymb"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));
               writer.println("  \\undef\\ggg");
               writer.println("  \\undef\\lll");
               writer.println("  \\undef\\checkmark");
               writer.println("  \\undef\\centerdot");
               writer.println("  \\undef\\veebar");
               writer.println("  \\undef\\barwedge");
               writer.println("  \\undef\\doublebarwedge");
               writer.println("  \\undef\\circledS");
               writer.println("  \\undef\\dotplus");
               writer.println("  \\undef\\complement");
               writer.println("  \\undef\\divideontimes");
               writer.println("  \\undef\\thicksim");
               writer.println("  \\undef\\thickapprox");
               writer.println("  \\undef\\hslash");

               pkgInfo.write(writer);

               writer.println("  \\let\\AMSSYMBggg\\ggg");
               writer.println("  \\let\\AMSSYMBlll\\lll");
               writer.println("  \\let\\AMSSYMBcheckmark\\checkmark");
               writer.println("  \\let\\AMSSYMBcenterdot\\centerdot");
               writer.println("  \\let\\AMSSYMBveebar\\veebar");
               writer.println("  \\let\\AMSSYMBbarwedge\\barwedge");
               writer.println("  \\let\\AMSSYMBdoublebarwedge\\doublebarwedge");
               writer.println("  \\let\\AMSSYMBcircledS\\circledS");
               writer.println("  \\let\\AMSSYMBdotplus\\dotplus");
               writer.println("  \\let\\AMSSYMBcomplement\\complement");
               writer.println("  \\let\\AMSSYMBdivideontimes\\divideontimes");
               writer.println("  \\let\\AMSSYMBthicksim\\thicksim");
               writer.println("  \\let\\AMSSYMBthickapprox\\thickapprox");
               writer.println("  \\let\\AMSSYMBhslash\\hslash");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoAMSSYMB{%");
               writer.println("   \\let\\ggg\\AMSSYMBggg");
               writer.println("   \\let\\lll\\AMSSYMBlll");
               writer.println("   \\let\\checkmark\\AMSSYMBcheckmark");
               writer.println("   \\let\\centerdot\\AMSSYMBcenterdot");
               writer.println("   \\let\\veebar\\AMSSYMBveebar");
               writer.println("   \\let\\barwedge\\AMSSYMBbarwedge");
               writer.println("   \\let\\doublebarwedge\\AMSSYMBdoublebarwedge");
               writer.println("   \\let\\circledS\\AMSSYMBcircledS");
               writer.println("   \\let\\dotplus\\AMSSYMBdotplus");
               writer.println("   \\let\\complement\\AMSSYMBcomplement");
               writer.println("   \\let\\divideontimes\\AMSSYMBdivideontimes");
               writer.println("   \\let\\thicksim\\AMSSYMBthicksim");
               writer.println("   \\let\\thickapprox\\AMSSYMBthickapprox");
               writer.println("   \\let\\hslash\\AMSSYMBhslash");
               writer.println("  }%");
            }
            else if (key.equals("stmaryrd"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));
               writer.println("  \\undef\\baro");
               writer.println("  \\undef\\lightning");
               writer.println("  \\let\\LATEXbigtriangledown\\bigtriangledown");
               writer.println("  \\let\\LATEXbigtriangleup\\bigtriangleup");

               pkgInfo.write(writer);

               writer.println("  \\let\\STMARYRDbaro\\baro");
               writer.println("  \\let\\STMARYRDlightning\\lightning");
               writer.println("  \\let\\STMARYRDbigtriangledown\\bigtriangledown");
               writer.println("  \\let\\STMARYRDbigtriangleup\\bigtriangleup");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoSTMARYRD{%");
               writer.println("   \\let\\baro\\STMARYRDbaro");
               writer.println("   \\let\\lightning\\STMARYRDlightning");
               writer.println("   \\let\\bigtriangledown\\STMARYRDbigtriangledown");
               writer.println("   \\let\\bigtriangleup\\STMARYRDbigtriangleup");
               writer.println("  }%");
               writer.println("  \\let\\bigtriangledown\\LATEXbigtriangledown");
               writer.println("  \\let\\bigtriangleup\\LATEXbigtriangleup");
            }
            else if (key.equals("wasysym"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\Circle");
               writer.println("  \\undef\\lightning");
               writer.println("  \\undef\\Square");

               pkgInfo.write(writer);

               writer.println("  \\let\\WASYSYMCircle\\Circle");
               writer.println("  \\let\\WASYSYMlightning\\lightning");
               writer.println("  \\let\\WASYSYMSquare\\Square");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoWASYSYM{%");
               writer.println("   \\let\\Circle\\WASYSYMCircle");
               writer.println("   \\let\\lightning\\WASYSYMlightning");
               writer.println("   \\let\\Square\\WASYSYMSquare");
               writer.println("  }%");
            }
            else if (key.equals("mathabx"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\ggg");
               writer.println("  \\undef\\lll");
               writer.println("  \\undef\\Sun");

               pkgInfo.write(writer);

               writer.println("  \\let\\MATHABXggg\\ggg");
               writer.println("  \\let\\MATHABXlll\\lll");
               writer.println("  \\let\\MATHABXSun\\Sun");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoMATHABX{%");
               writer.println("   \\let\\ggg\\MATHABXggg");
               writer.println("   \\let\\lll\\MATHABXlll");
               writer.println("   \\let\\Sun\\MATHABXSun");
               writer.println("  }%");
            }
            else if (key.equals("marvosym"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\Cross");
               writer.println("  \\undef\\Letter");
               writer.println("  \\undef\\Lightning");
               writer.println("  \\undef\\Sun");
               writer.println("  \\undef\\Email");

               pkgInfo.write(writer);

               writer.println("  \\let\\MARVOSYMCross\\Cross");
               writer.println("  \\let\\MARVOSYMLetter\\Letter");
               writer.println("  \\let\\MARVOSYMLightning\\Lightning");
               writer.println("  \\let\\MARVOSYMSun\\Sun");
               writer.println("  \\let\\MARVOSYMEmail\\Email");
               writer.println("  \\undef\\Email");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoMARVOSYM{%");
               writer.println("   \\let\\Cross\\MARVOSYMCross");
               writer.println("   \\let\\Letter\\MARVOSYMLetter");
               writer.println("   \\let\\Lightning\\MARVOSYMLightning");
               writer.println("   \\let\\Sun\\MARVOSYMSun");
               writer.println("  }%");
            }
            else if (key.equals("bbding"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\Cross");
               writer.println("  \\undef\\Square");
               writer.println("  \\undef\\TriangleDown");
               writer.println("  \\undef\\TriangleUp");

               pkgInfo.write(writer);

               writer.println("  \\let\\BBDINGCross\\Cross");
               writer.println("  \\let\\BBDINGSquare\\Square");
               writer.println("  \\let\\BBDINGTriangleDown\\TriangleDown");
               writer.println("  \\let\\BBDINGTriangleUp\\TriangleUp");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoBBDING{%");
               writer.println("   \\let\\Cross\\BBDINGCross");
               writer.println("   \\let\\Square\\BBDINGSquare");
               writer.println("   \\let\\TriangleDown\\BBDINGTriangleDown");
               writer.println("   \\let\\TriangleUp\\BBDINGTriangleUp");
               writer.println("  }%");
            }
            else if (key.equals("ifsym"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\Circle");
               writer.println("  \\undef\\Cross");
               writer.println("  \\undef\\Letter");
               writer.println("  \\undef\\Lightning");
               writer.println("  \\undef\\Square");
               writer.println("  \\undef\\Sun");
               writer.println("  \\undef\\TriangleDown");
               writer.println("  \\undef\\TriangleUp");

               pkgInfo.write(writer);

               writer.println("  \\let\\IFSYMCircle\\Circle");
               writer.println("  \\let\\IFSYMCross\\Cross");
               writer.println("  \\let\\IFSYMLetter\\Letter");
               writer.println("  \\let\\IFSYMLightning\\Lightning");
               writer.println("  \\let\\IFSYMSquare\\Square");
               writer.println("  \\let\\IFSYMSun\\Sun");
               writer.println("  \\let\\IFSYMTriangleDown\\TriangleDown");
               writer.println("  \\let\\IFSYMTriangleUp\\TriangleUp");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoIFSYM{%");
               writer.println("   \\let\\Circle\\IFSYMCircle");
               writer.println("   \\let\\Cross\\IFSYMCross");
               writer.println("   \\let\\Letter\\IFSYMLetter");
               writer.println("   \\let\\Lightning\\IFSYMLightning");
               writer.println("   \\let\\Square\\IFSYMSquare");
               writer.println("   \\let\\Sun\\IFSYMSun");
               writer.println("   \\let\\TriangleDown\\IFSYMTriangleDown");
               writer.println("   \\let\\TriangleUp\\IFSYMTriangleUp");
               writer.println("  }%");
            }
            else if (key.equals("dingbat"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\checkmark");

               pkgInfo.write(writer);

               writer.println("  \\let\\DINGBATcheckmark\\checkmark");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoDINGBAT{%");
               writer.println("   \\let\\checkmark\\DINGBATcheckmark");
               writer.println("  }%");
            }
            else if (key.equals("wsuipa"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\baro");

               pkgInfo.write(writer);

               writer.println("  \\let\\WSUIPAbaro\\baro");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoWSUIPA{%");
               writer.println("   \\let\\baro\\WSUIPAbaro");
               writer.println("  }%");
            }
            else if (key.equals("MnSymbol"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\ggg");
               writer.println("  \\undef\\lll");
               writer.println("  \\undef\\checkmark");
               writer.println("  \\undef\\centerdot");
               writer.println("  \\undef\\veebar");
               writer.println("  \\undef\\barwedge");
               writer.println("  \\undef\\doublebarwedge");
               writer.println("  \\undef\\circledS");
               writer.println("  \\undef\\dotplus");
               writer.println("  \\undef\\complement");
               writer.println("  \\undef\\divideontimes");
               writer.println("  \\undef\\thicksim");
               writer.println("  \\undef\\thickapprox");
               writer.println("  \\undef\\hslash");

               pkgInfo.write(writer);

               writer.println("  \\let\\MNSYMBOLggg\\ggg");
               writer.println("  \\let\\MNSYMBOLlll\\lll");
               writer.println("  \\let\\MNSYMBOLcheckmark\\checkmark");
               writer.println("  \\let\\MNSYMBOLcenterdot\\centerdot");
               writer.println("  \\let\\MNSYMBOLveebar\\veebar");
               writer.println("  \\let\\MNSYMBOLbarwedge\\barwedge");
               writer.println("  \\let\\MNSYMBOLdoublebarwedge\\doublebarwedge");
               writer.println("  \\let\\MNSYMBOLcircledS\\circledS");
               writer.println("  \\let\\MNSYMBOLdotplus\\dotplus");
               writer.println("  \\let\\MNSYMBOLcomplement\\complement");
               writer.println("  \\let\\MNSYMBOLdivideontimes\\divideontimes");
               writer.println("  \\let\\MNSYMBOLthicksim\\thicksim");
               writer.println("  \\let\\MNSYMBOLthickapprox\\thickapprox");
               writer.println("  \\let\\MNSYMBOLhslash\\hslash");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoMNSYMBOL{%");
               writer.println("   \\let\\ggg\\MNSYMBOLggg");
               writer.println("   \\let\\lll\\MNSYMBOLlll");
               writer.println("   \\let\\checkmark\\MNSYMBOLcheckmark");
               writer.println("   \\let\\centerdot\\MNSYMBOLcenterdot");
               writer.println("   \\let\\veebar\\MNSYMBOLveebar");
               writer.println("   \\let\\barwedge\\MNSYMBOLbarwedge");
               writer.println("   \\let\\doublebarwedge\\MNSYMBOLdoublebarwedge");
               writer.println("   \\let\\circledS\\MNSYMBOLcircledS");
               writer.println("   \\let\\dotplus\\MNSYMBOLdotplus");
               writer.println("   \\let\\complement\\MNSYMBOLcomplement");
               writer.println("   \\let\\divideontimes\\MNSYMBOLdivideontimes");
               writer.println("   \\let\\thicksim\\MNSYMBOLthicksim");
               writer.println("   \\let\\thickapprox\\MNSYMBOLthickapprox");
               writer.println("   \\let\\hslash\\MNSYMBOLhslash");
               writer.println("  }%");
            }
            else
            {
               writer.print(" ");

               if (!pkgInfo.isInstalled())
               {
                  writer.println("% "+MakeJmlrBookGUI.getLabelWithValue(
                    "newbook.comment.not_installed", key));
                  writer.print("%");
               }

               pkgInfo.write(writer);
            }
         }

         writer.println("}");

         writer.println();

         boolean isCiML = dialog.isCiML();

         writer.println("% "+MakeJmlrBookGUI.getLabelWithValues(
           "newbook.comment.generated", MakeJmlrBookGUI.appName,
            MakeJmlrBookGUI.appVersion));

         writer.println("% "
           +MakeJmlrBookGUI.getLabel("newbook.comment.wcporpmlr"));

         if (isCiML)
         {
            writer.println("% "
              +MakeJmlrBookGUI.getLabel("newbook.comment.ciml"));
            writer.println("\\documentclass[pmlr,7x10]{jmlrbook}");
         }
         else
         {
            writer.println("% "
              +MakeJmlrBookGUI.getLabel("newbook.comment.jmlrwcp"));
            writer.println("\\documentclass[pmlr]{jmlrbook}");
         }

         writer.println();
         writer.println("\\input{bookfonts}");
         writer.println();

         if (isCiML)
         {
            writer.println("% "
               +MakeJmlrBookGUI.getLabel("newbook.comment.title"));
         }

         writer.print("\\title");

         if (!shortTitle.isEmpty())
         {
            writer.print("["+shortTitle+"]");
         }

         writer.print("{");
         writer.print(title);
         writer.println("}");
         writer.println();

         writer.print("\\author");

         int n = editors.size();
         Contributor firstEd = editors.firstElement();

         if (n > 1)
         {
            writer.print(String.format("[%s et al.]", abbrvFormat.format(firstEd)));
         }

         String surname = firstEd.getSurname();

         if (surname.contains(" "))
         {
            surname = "{"+surname+"}";
         }

         writer.print("{"+firstEd.getForenames()+" "+surname);

         for (int i = 1; i < n; i++)
         {
            writer.print(i == n-1 ? " and " : ", ");

            Contributor contrib = editors.get(i);

            surname = contrib.getSurname();

            if (surname.contains(" "))
            {
               surname = "{"+surname+"}";
            }

            writer.print(contrib.getForenames()+" "+surname);
         }

         writer.println("}");

         if (isCiML)
         {
            writer.println();
            writer.println("\\usepackage[utf8]{inputenc}");
            writer.println("\\usepackage[T1]{fontenc}");

            writer.println();
            writer.println("\\ifprint{}{\\usepackage{bookmark}}% "
             + MakeJmlrBookGUI.getLabel("newbook.comment.load_last"));
            writer.println();

            writer.println("\\subtitle{"+dialog.getBookSubTitle()+"}");
            writer.println("\\volume{"+dialog.getVolume()+"}");
            writer.println("%\\logo{\\includegraphics{ciml-logo}}");
            writer.println("\\team{}");
            writer.println("\\productioneditor{"
              +dialog.getProductionEditor()+"}");

            writer.println();

            int wcpVol = dialog.getWCPVolume();

            if (wcpVol == 0)
            {
               writer.println("% "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.jmlrvolume", "\\jmlrvolume{n}"));
               writer.println("%\\jmlrvolume{0}");
            }
            else
            {
               writer.println("\\jmlrvolume{"+wcpVol+"}");
            }
         }
         else
         {
            writer.println("\\usepackage[utf8]{inputenc}");
            writer.println("\\usepackage[T1]{fontenc}");

            writer.println();
            writer.println("\\ifprint{}{\\usepackage{bookmark}}% "
             + MakeJmlrBookGUI.getLabel("newbook.comment.load_last"));
            writer.println();

            writer.println("\\jmlrvolume{"+dialog.getVolume()+"}");

            int year = dialog.getYear();

            writer.println("\\jmlryear{"+year+"}");

            writer.println();

            String month = dialog.getMonth();
            String day = dialog.getDay();

            writer.printf("\\date{%s, %d}%n", (day.isEmpty() ? month :
             month+" "+day), year);

            String location = dialog.getJMLRLocation();

            if (location != null)
            {
               writer.println("\\jmlrlocation{"+location+"}");
            }

            writer.println();

            writer.println("% "+MakeJmlrBookGUI.getLabelWithValue(
              "newbook.comment.imagefile", "imageFile"));
            writer.println("% \\logo[url]{\\includegraphics{imageFile}}");
         }

         writer.println();
         writer.println("\\begin{document}");
         writer.println("\\maketitle");

         if (isCiML)
         {
            writer.println("\\frontmatter");

            if (hasForeword)
            {
               writeInputCs(bookPath, forewordFile);
            }

            if (hasPreface)
            {
               writeInputCs(bookPath, prefaceFile);
            }

            writer.println("\\tableofcontents");

            writer.println("\\mainmatter");
         }
         else
         {
            writer.println("\\frontmatter");
            writer.println("\\tableofcontents");

            writer.println("\\mainmatter");

            if (hasPreface)
            {
               writeInputCs(bookPath, prefaceFile);
            }
         }

         writeMainMatter(bookPath);

         if (hasAppendices)
         {
            writeAppendices(bookPath);
         }

         writer.println("\\end{document}");
      }
      catch (IOException e)
      {
         JMLRResources.error(dialog, e);
      }
      catch (Throwable e)
      {
         MakeJmlrBookGUI.debug(e);
      }
      finally
      {
         if (writer != null)
         {
            writer.close();
            writer = null;
         }
      }
   }

   protected void writeInputCs(Path basePath, File file)
     throws IOException
   {
      String path = MakeJmlrBookGUI.getTeXPath(basePath, file.toPath());

      if (path.endsWith(".tex"))
      {
         path = path.substring(0, path.length()-4);
      }

      writer.println("\\input{"+path+"}");
   }

   protected void writeMainMatter(Path bookPath)
     throws IOException
   {
      if (partConfig == BookPartPanel.PARTS_NO)
      {
         ImportPapersPanel mainImportPanel = dialog.getMainImportPapers();

         writePapers(bookPath, mainImportPanel.getPapers());
      }
      else
      {
         Vector<PartListPanel> partPanels = dialog.getPartLists();

         for (PartListPanel partPanel : partPanels)
         {
            if (partConfig == BookPartPanel.PARTS_TOC)
            {
               writer.print("\\addtocpart");
            }
            else
            {
               writer.print("\\part");
            }

            writer.println("{"+partPanel.getPartTitle()+"}");

            ImportPapersPanel impPanel = partPanel.getImportPapersPanel();

            writePapers(bookPath, impPanel.getPapers());
         }
      }
   }

   protected void writeAppendices(Path bookPath)
     throws IOException
   {
      writer.println();
      writer.println("\\appendix");

      ImportPapersPanel impPanel = dialog.getAppendixImportPapers();

      writePapers(bookPath, impPanel.getPapers());
   }

   protected void writePapers(Path bookPath, Vector<ImportPaperPanel> papers)
     throws IOException
   {
      writer.println();
      writer.println("\\begin{jmlrpapers}");

      for (ImportPaperPanel paper: papers)
      {
         ImportFileAttributes attr = paper.getAttributes();

         if (attr == null)
         {
            attr = attrMap.get(paper.getFile());
            paper.setAttributes(attr);
         }

         File file = paper.getToImportFile();

         if (file == null)
         {
            continue;
         }

         String parent = MakeJmlrBookGUI.getTeXPath(bookPath,
           file.getParentFile().toPath());

         if (parent == null) parent = ".";

         String base = file.getName();

         if (base.endsWith(".tex"))
         {
            base = base.substring(0, base.length()-4);
         }

         if (paper.isReprint())
         {
            writer.print("\\importpubpaper{");
            writer.print(parent+"}{");
            writer.print(base);
            writer.print("}{");
            writer.print(paper.getStartPage());
            writer.print("--");
            writer.print(paper.getEndPage());
            writer.println("}");
         }
         else if (dialog.isCiML())
         {
            writer.print("\\importarticle{");
            writer.print(parent+"}{");
            writer.print(base);
            writer.println("}");
         }
         else
         {
            writer.print("\\importpaper{");
            writer.print(parent+"}{");
            writer.print(base);
            writer.println("}");
         }
      }

      writer.println("\\end{jmlrpapers}");
      writer.println();
   }

   protected void parsing(File file)
   {
      message(MakeJmlrBookGUI.getLabelWithValue("message.parsing",
        file.getAbsolutePath()));
   }

   protected String tag(String text)
   {
      return "<b><tt>"+text+"</tt></b>";
   }

   protected void message(String tag, String msg)
   {
      messageArea.update("<"+tag+">"+msg+"</"+tag+">");
   }

   public void message(String msg)
   {
      message("p", msg);
   }

   public String getMessage(String label)
   {
      return MakeJmlrBookGUI.getLabel(label);
   }

   public String getMessage(String label, String param)
   {
      return MakeJmlrBookGUI.getLabelWithValue(label, param);
   }

   public String getMessage(String label, String[] params)
   {
      return MakeJmlrBookGUI.getLabelWithValues(label, params);
   }

   public String getMessage(String label, Object... params)
   {
      return MakeJmlrBookGUI.getLabelWithValues(label, params);
   }

   protected void error(String msg)
   {
      errMessageBuffer.append('\n');

      if (currentFile == null)
      {
         errMessageBuffer.append(msg.replaceAll("<[^>]+>", ""));
      }
      else
      {
         errMessageBuffer.append(currentFile.getAbsolutePath()
           +": "+msg.replaceAll("<[^>]+>", ""));
      }

      messageArea.update("<p><font color=\"red\">"+msg+"</font></p>");
   }

   public void warning(TeXParser parser, String msg)
   {
      File file = parser.getCurrentFile();
      int lineNum = parser.getLineNumber();

      if (lineNum > 0)
      {
         msg = MakeJmlrBookGUI.getLabelWithValues("error.line", ""+lineNum, msg);

         if (file != null)
         {
            msg = MakeJmlrBookGUI.getLabelWithValues("error.file",
              file.getName(), msg);
         }
      }

      messageArea.update("<p><font color=\"red\">"+msg+"</font></p>");
   }

   public void substituting(TeXParser parser, String original, String replacement)
   {
      String msg;

      if (replacement.isEmpty())
      {
         msg = MakeJmlrBookGUI.getLabelWithValue("warning.removing", 
             tag(original));
      }
      else
      {
         msg = MakeJmlrBookGUI.getLabelWithValues("warning.substituting",
              tag(original), tag(replacement));
      }

      int lineNumber = parser.getLineNumber();

      if (lineNumber > 0)
      {
         msg = MakeJmlrBookGUI.getLabelWithValues("error.line",
          ""+lineNumber, msg);
      }

      File file = parser.getCurrentFile();

      if (file != null)
      {
         msg = MakeJmlrBookGUI.getLabelWithValues("error.file",
           file.getName(), msg);
      }

      message(msg);
   }

   public void error(Exception e)
   {
      if (e instanceof TeXSyntaxException)
      {
         error(((TeXSyntaxException)e).getMessage(this));
      }
      else if (e instanceof NewBookException)
      {
         error(((NewBookException)e).getDetails());
      }
      else
      {
         error(e.getMessage());
      }
   }


   protected File getPackageLocation(String name)
   {
      return dialog.getPackageLocation(name);
   }

   public boolean isProblemSty(String name)
   {
      for (int i = 0; i < PROBLEM_STYS.length; i++)
      {
         if (PROBLEM_STYS[i].equals(name)) return true;
      }

      return false;
   }

   public boolean isIgnoreSty(String name)
   {
      for (int i = 0; i < IGNORE_STYS.length; i++)
      {
         if (IGNORE_STYS[i].equals(name)) return true;
      }

      return false;
   }

   public File getBookDir()
   {
      return bookDir;
   }

   private NewBookDialog dialog;

   private MessagePane messageArea;

   private JPanel buttonPanel;

   private StringBuffer errMessageBuffer;

   private Hashtable<String,PackageInfo> styNames;

   private Hashtable<File,ImportFileAttributes> attrMap;

   public static final String[] PROBLEM_STYS = new String[]
   {
      "epsfig", "psfig", "subfig", "theorem", "ntheorem",
      "amsthm", "pdfpages", "geometry", "tabularx", "algorithm"
   };

   // Packages that can be skipped because they're already loaded by
   // jmlr.cls

   public static final String[] IGNORE_STYS = new String[]
   {
      "url", "amsmath", "amssymb", "natbib", "graphicx", "xcolor",
      "algorithm2e", "hyperref", "nameref", "etoolbox", "calc",
      "xkeyval", "setspace", "bookmark"
   };

   private Hashtable<File,File> importFiles;

   private BufferedReader reader;
   private PrintWriter writer;
   private File currentFile;
   private File bookDir, bookFile;
   private String base;
   private File prefaceFile, forewordFile, frontMatterDir;
   private int partConfig;

   private ContributorFormat contribFormat = new PlainContributorFormat();
   private ContributorFormat abbrvFormat = new AbbrvContributorFormat();

   private String title, shortTitle;

   private Vector<Contributor> editors;

   private boolean hasPreface, hasForeword, hasAppendices;

   private static final String FRONTMATTER="FrontMatter";
   private static final String PREFACE="preface";
   private static final String FOREWORD="foreword";
}
